"""Unit tests for the Clease calculator."""
import os
from random import randint
import numpy as np
from clease import CEBulk, CECrystal, CorrFunction, Concentration
from clease.calculator import Clease, attach_calculator
from ase.build import bulk
from ase.spacegroup import crystal
from clease.tools import wrap_and_sort_by_position
import time
import unittest


def generate_ex_eci(setting):
    """Return dummy ECIs. All are set to -0.001."""
    cf = CorrFunction(setting)
    cf = cf.get_cf(setting.atoms)
    eci = {key: -0.001 for key in cf}
    return eci


def get_binary(db_name):
    """Return a simple binary test structure."""
    basis_elements = [["Au", "Cu"]]
    concentration = Concentration(basis_elements=basis_elements)
    bc_setting = CEBulk(crystalstructure="fcc",
                        a=4.05,
                        size=[3, 3, 3],
                        concentration=concentration,
                        db_name=db_name,
                        max_cluster_size=3,
                        max_cluster_dia=[5.0, 5.0])

    atoms = bulk("Au", crystalstructure="fcc", a=4.05)
    atoms = atoms * (3, 3, 3)
    for i in range(int(len(atoms) / 2)):
        atoms[i].symbol = "Au"
        atoms[-i - 1].symbol = "Cu"
    return bc_setting, wrap_and_sort_by_position(atoms)


def get_ternary(db_name):
    """Return a ternary test structure."""
    basis_elements = [["Au", "Cu", "Zn"]]
    concentration = Concentration(basis_elements=basis_elements)
    bc_setting = CEBulk(crystalstructure="fcc",
                        a=4.05,
                        size=[3, 3, 3],
                        concentration=concentration,
                        db_name=db_name,
                        max_cluster_dia=[5.0, 5.0],
                        max_cluster_size=3)

    atoms = bulk("Au", crystalstructure="fcc", a=4.05)
    atoms = atoms * (3, 3, 3)
    for i in range(2):
        atoms[3 * i].symbol = "Au"
        atoms[3 * i + 1].symbol = "Cu"
        atoms[3 * i + 2].symbol = "Zn"
    return bc_setting, wrap_and_sort_by_position(atoms)


def get_rocksalt(db_name):
    """Test rocksalt where passed atoms with background_atoms."""
    basis_elements = [['Li', 'X', 'V'], ['O']]
    concentration = Concentration(basis_elements=basis_elements)
    setting = CEBulk(crystalstructure='rocksalt',
                     a=4.05,
                     size=[3, 3, 3],
                     concentration=concentration,
                     db_name=db_name,
                     max_cluster_size=3,
                     max_cluster_dia=[7.0, 7.0],
                     ignore_background_atoms=True)

    atoms = bulk("LiO", crystalstructure="rocksalt", a=4.05)
    atoms = atoms * (3, 3, 3)
    Li_indx = [a.index for a in atoms if a.symbol == 'Li']
    for i in range(18):
        if i < 9:
            atoms[Li_indx[i]].symbol = 'V'
        else:
            atoms[Li_indx[i]].symbol = 'X'
    return setting, wrap_and_sort_by_position(atoms)


def rocksalt_with_self_interaction(size, db_name):
    basis_elements = [['Li', 'Mn', 'X'], ['O', 'X']]
    concentration = Concentration(basis_elements=basis_elements)
    setting = CEBulk(crystalstructure='rocksalt',
                     a=4.05,
                     size=size,
                     concentration=concentration,
                     db_name=db_name,
                     max_cluster_size=3,
                     basis_function='trigonometric',
                     max_cluster_dia=[7.0, 4.0])
    atoms = setting.atoms.copy()
    return setting, atoms


def get_spacegroup(db_name):
    """Test rocksalt where passed atoms."""
    basis = [(0., 0., 0.),
             (0.3894, 0.1405, 0.),
             (0.201, 0.3461, 0.5),
             (0.2244, 0.3821, 0.)]
    spacegroup = 55
    cellpar = [6.25, 7.4, 3.83, 90, 90, 90]
    size = [2, 2, 2]
    basis_elements = [['O', 'X'], ['O', 'X'], ['O', 'X'], ['Ta']]
    grouped_basis = [[0, 1, 2], [3]]
    concentration = Concentration(basis_elements=basis_elements,
                                  grouped_basis=grouped_basis)

    setting = CECrystal(basis=basis,
                        spacegroup=spacegroup,
                        cellpar=cellpar,
                        size=size,
                        concentration=concentration,
                        db_name=db_name,
                        max_cluster_size=3,
                        max_cluster_dia=[5.0, 5.0],
                        ignore_background_atoms=False)

    atoms = crystal(symbols=['O', 'X', 'O', 'Ta'], basis=basis,
                    spacegroup=spacegroup, cell=None,
                    cellpar=cellpar, ab_normal=(0, 0, 1),
                    size=size)

    return setting, wrap_and_sort_by_position(atoms)


def test_update_correlation_functions(setting, atoms, n_trial_configs=20,
                                      fixed=[]):
    """Perform swaps and check that the correlation functions match.

    The comparison is done by check that each CF in the Clease
    calculator is the same as the ones obtained by direct calculation.
    """
    cf = CorrFunction(setting)

    eci = generate_ex_eci(setting)
    calc = Clease(setting, eci=eci)
    atoms.set_calculator(calc)

    timings = []
    for _ in range(n_trial_configs):
        indx1 = randint(0, len(atoms) - 1)
        symb1 = atoms[indx1].symbol
        while symb1 in fixed:
            indx1 = randint(0, len(atoms) - 1)
            symb1 = atoms[indx1].symbol
        symb2 = symb1
        while symb2 == symb1 or symb2 in fixed:
            indx2 = randint(0, len(atoms) - 1)
            symb2 = atoms[indx2].symbol

        atoms[indx1].symbol = symb2
        atoms[indx2].symbol = symb1

        # The calculator should update its correlation functions
        # when the energy is computed
        start = time.time()
        atoms.get_potential_energy()
        timings.append(time.time() - start)
        brute_force_cf = cf.get_cf_by_names(atoms, calc.cf_names)
        calc_cf = calc.get_cf()

        for key in calc_cf.keys():
            assert abs(calc_cf[key] - brute_force_cf[key]) < 1E-6

    print(np.mean(timings))


def test_insert_element(setting, atoms, n_trial_configs=20):
    from random import choice
    cf = CorrFunction(setting)
    eci = generate_ex_eci(setting)
    calc = Clease(setting, eci=eci)
    atoms.set_calculator(calc)
    elements = setting.unique_elements
    for _ in range(n_trial_configs):
        indx1 = randint(0, len(atoms)-1)
        symb1 = atoms[indx1].symbol
        symb2 = symb1

        while symb2 == symb1:
            symb2 = choice(elements)
        atoms[indx1].symbol = symb2
        atoms.get_potential_energy()
        brute_force_cf = cf.get_cf_by_names(atoms, calc.cf_names)
        calc_cf = calc.get_cf()
        for k in calc_cf.keys():
            if k.startswith("c0") or k.startswith("c1"):
                continue
            assert abs(calc_cf[k] - brute_force_cf[k]) < 1E-6


class TestCECalculator(unittest.TestCase):
    def test_normfactors_no_self_interaction(self):
        db_name = 'cecalc_binary_norm_fac.db'
        setting, atoms = get_binary(db_name)

        eci = generate_ex_eci(setting)
        calc = Clease(setting, eci=eci)
        atoms.set_calculator(calc)

        for cluster in setting.cluster_list:
            if cluster.name == 'c0' or cluster.name == 'c1':
                continue
            norm_factors = cluster.info['normalization_factor']
            self.assertTrue(np.allclose(norm_factors, 1.0))


    def test_indices_of_changed_symbols(self):
        db_name = 'indices_changes_symbol.db'
        setting, atoms = get_binary(db_name)
        eci = generate_ex_eci(setting)
        calc = Clease(setting, eci=eci)
        atoms.set_calculator(calc)

        changes = [2, 6]
        for ch in changes:
            if atoms[ch].symbol == 'Au':
                atoms[ch].symbol = 'Cu'
            else:
                atoms[ch].symbol = 'Au'

        calc_changes = calc.indices_of_changed_atoms
        os.remove(db_name)
        self.assertEqual(calc_changes, changes)

    def test_update_corr_func_binary(self):
        db_name = 'cecalc_corr_func_binary.db'
        print('binary')
        bin_setting, bin_atoms = get_binary(db_name)
        test_update_correlation_functions(bin_setting, bin_atoms,
                                          n_trial_configs=5)
        os.remove(db_name)

    def test_update_corr_func_ternary(self):
        db_name = 'cecalc_corr_func_ternary.db'
        print('ternary')
        tern_setting, tern_atoms = get_ternary(db_name)
        test_update_correlation_functions(tern_setting, tern_atoms,
                                          n_trial_configs=5)
        os.remove(db_name)

    def test_update_corr_func_rocksalt(self):
        db_name = 'cecalc_corr_func_rocksalt.db'
        print('rocksalt')
        rs_setting, rs_atoms = get_rocksalt(db_name)
        test_update_correlation_functions(rs_setting, rs_atoms,
                                          n_trial_configs=5, fixed=['O'])
        os.remove(db_name)


    def test_insert_element_rocksalt_1x1x1(self):
        print('rocksalt with self interaction 1x1x1')
        db_name = 'cecalc_rs_1x1x1.db'
        rs_setting, rs_atoms = rocksalt_with_self_interaction([1, 1, 1],
                                                              db_name)
        test_insert_element(rs_setting, rs_atoms, n_trial_configs=5)
        os.remove(db_name)

    def test_insert_element_rocksalt_1x1x2(self):
        db_name = 'cecalc_rs_1x1x2.db'
        print('rocksalt with self interaction 1x1x2')
        rs_setting, rs_atoms = rocksalt_with_self_interaction([1, 1, 2],
                                                              db_name)
        test_insert_element(rs_setting, rs_atoms, n_trial_configs=1)
        os.remove(db_name)


    def test_insert_element_rocksalt_1x1x3(self):
        db_name = 'cecalc_rs_1x1x3.db'
        print('rocksalt with self interaction 1x1x3')
        rs_setting, rs_atoms = rocksalt_with_self_interaction([1, 1, 3],
                                                              db_name)
        test_insert_element(rs_setting, rs_atoms, n_trial_configs=10)
        os.remove(db_name)


    def test_insert_element_rocksalt_1x2x3(self):
        print('rocksalt with self interaction 1x2x3')
        db_name = 'cecalc_rs_1x2x3.db'
        rs_setting, rs_atoms = rocksalt_with_self_interaction([1, 2, 3],
                                                              db_name)
        test_insert_element(rs_setting, rs_atoms, n_trial_configs=10)
        os.remove(db_name)

    def test_update_corr_func_spacegroup(self):
        print('spacegroup')
        db_name = 'cecalc_corrfunc_spacegroup.db'
        sp_setting, sp_atoms = get_spacegroup(db_name)
        test_update_correlation_functions(sp_setting, sp_atoms,
                                          n_trial_configs=5, fixed=['Ta'])
        os.remove(db_name)

    def test_init_large_cell(self):
        print('Init large cell')
        db_name = 'cecalc_init_large_cell.db'
        rs_setting, _ = rocksalt_with_self_interaction([1, 2, 3], db_name)

        atoms = bulk('LiO', crystalstructure='rocksalt', a=4.05, cubic=True)
        atoms = atoms*(2, 2, 2)
        eci = generate_ex_eci(rs_setting)

        # Use quick way of initialisation object
        atoms = attach_calculator(setting=rs_setting, atoms=atoms, eci=eci)

        cf = CorrFunction(rs_setting)
        init_cf = atoms.get_calculator().init_cf

        final_cf = cf.get_cf(atoms)
        for k, v in final_cf.items():
            self.assertAlmostEqual(v, init_cf[k])

        # Try some swaps
        num_X = 0
        num_Mn = 0
        for atom in atoms:
            if atom.symbol == 'Li' and num_X < 3:
                atom.symbol = 'X'
                num_X += 1
            elif atom.symbol == 'Li' and num_Mn < 4:
                atom.symbol = 'Mn'
                num_Mn += 1
        atoms.get_potential_energy()

        final_cf = cf.get_cf(atoms)
        calc_cf = atoms.get_calculator().get_cf()
        os.remove(db_name)
        for k, v in final_cf.items():
            self.assertAlmostEqual(v, calc_cf[k])

    def test_4body_attach(self):
        conc = Concentration(basis_elements=[['Au', 'Cu']])
        setting = CEBulk(crystalstructure='fcc',
                         a=4.0,
                         size=[2, 2, 2],
                         concentration=conc,
                         db_name="test_aucu.db",
                         max_cluster_size=4,
                         max_cluster_dia=[6.0, 5.0, 5.0],
                         basis_function='polynomial')

        cf = CorrFunction(setting).get_cf(setting.atoms)
        eci = {k: 0.0 for k in cf.keys()}
        eci['c0'] = 1.0
        atoms = setting.atoms.copy()*(3, 3, 3)

        # Simpy confirm that no exception is raised.
        # In the past, this failed.
        _ = attach_calculator(setting, atoms=atoms, eci=eci)
        os.remove('test_aucu.db')

    def test_given_change_and_restore(self):
        db_name = 'test_given_change.db'
        setting, atoms = get_binary(db_name)

        for atom in atoms:
            atom.symbol = 'Au'

        calc = Clease(setting, eci=generate_ex_eci(setting))
        atoms.set_calculator(calc)

        os.remove(db_name)
        cf = CorrFunction(setting)

        init_cf = cf.get_cf(atoms)

        # Insert to Cu atoms
        _ = atoms.get_calculator().get_energy_given_change(
            [(0, 'Au', 'Cu'), (1, 'Au', 'Cu')])

        # We should have to Cu atoms now
        num_cu = sum(1 for atom in atoms if atom.symbol == 'Cu')
        self.assertEqual(num_cu, 2)

        cf_calc_two_inserts = atoms.get_calculator().get_cf()
        cf_scratch = cf.get_cf(atoms)

        for k, v in cf_calc_two_inserts.items():
            self.assertAlmostEqual(v, cf_scratch[k])

        atoms.get_calculator().restore()

        # Now we should be back to pure Au
        num_au = sum(1 for atom in atoms if atom.symbol == 'Au')
        self.assertEqual(num_au, len(atoms))

        cf_calc = calc.get_cf()
        for k, v in cf_calc.items():
            self.assertAlmostEqual(v, init_cf[k])

        # Insert two atoms again
        _ = atoms.get_calculator().get_energy_given_change(
            [(0, 'Au', 'Cu'), (1, 'Au', 'Cu')]
        )

        # Clear the history
        atoms.get_calculator().clear_history()

        # Restore should now not have any effect
        atoms.get_calculator().restore()
        cf_calc = calc.get_cf()

        # Should still be two Cu atoms
        num_cu = sum(1 for atom in atoms if atom.symbol == 'Cu')
        self.assertEqual(num_cu, 2)

        for k, v in cf_calc_two_inserts.items():
            self.assertAlmostEqual(cf_calc[k], v)
        os.remove(db_name)


if __name__ == '__main__':
    unittest.main()



import unittest
from ase.build import bulk
from clease.tools import min_distance_from_facet, factorize
from clease.tools import all_integer_transform_matrices


class TestTools(unittest.TestCase):
    def test_min_distance_from_facet(self):
        a = 4.0
        atoms = bulk('Al', crystalstructure='sc', a=a)

        x = [3.0, 3.5, 3.9]
        dist = min_distance_from_facet(x, atoms.get_cell())
        self.assertAlmostEqual(dist, 0.1)

    def test_factorize(self):
        fact = sorted(list(factorize(10)))
        self.assertEqual(fact, [2, 5])

        fact = sorted(list(factorize(16)))
        self.assertEqual(fact, [2, 2, 2, 2])

        fact = sorted(list(factorize(24)))
        self.assertEqual(fact, [2, 2, 2, 3])

    def test_all_int_matrices(self):
        _ = all_integer_transform_matrices(10)


if __name__ == '__main__':
    unittest.main()



from ase.build import bulk
from ase.spacegroup import crystal
from clease.trans_matrix_constructor import TransMatrixConstructor
from clease.tools import wrap_and_sort_by_position, index_by_position
import unittest


def brute_force_tm_construction(ref_indices, index_by_basis, atoms):

    tm = [[] for i in range(len(atoms))]
    for ref, indices in zip(ref_indices, index_by_basis):
        for i in indices:
            vec = atoms.get_distance(i, ref, vector=True)
            shifted = atoms.copy()
            shifted.translate(vec)
            shifted.wrap()
            tm[i] = index_by_position(shifted)
    return tm


def check_sparse_dense(sparse, dense):
    for i, row in enumerate(dense):
        sp_row = sparse[i]
        for col, val in sp_row.items():
            if val != row[col]:
                return False
    return True


class TestTransMatrixConstructor(unittest.TestCase):
    def test_fcc(self):
        atoms = bulk("Al", crystalstructure="fcc")*(4, 4, 4)
        atoms = wrap_and_sort_by_position(atoms)

        tm_constructor = TransMatrixConstructor(atoms, 5.0)

        ref_indx = [0]
        indx_by_basis = [list(range(len(atoms)))]
        tm_brute = brute_force_tm_construction(ref_indx, indx_by_basis, atoms)
        symm_group = [0]*len(atoms)

        tm_fast = tm_constructor.construct(ref_indx, symm_group)
        self.assertTrue(check_sparse_dense(tm_fast, tm_brute))

    def test_two_basis(self):
        atoms = bulk("NaCl", crystalstructure="rocksalt", a=5.0)
        atoms[0].tag = 0
        atoms[1].tag = 1
        atoms = atoms*(3, 4, 5)
        atoms = wrap_and_sort_by_position(atoms)

        symm_group = [atom.tag for atom in atoms]
        index_by_group = [[], []]
        for atom in atoms:
            index_by_group[atom.tag].append(atom.index)
        ref_indx = [min(index_by_group[0]), min(index_by_group[1])]

        tm_brute = brute_force_tm_construction(ref_indx, index_by_group, atoms)

        tm_constructor = TransMatrixConstructor(atoms, 5.5)
        tm_fast = tm_constructor.construct(ref_indx, symm_group)
        self.assertTrue(check_sparse_dense(tm_fast, tm_brute))

    def test_four_basis(self):
        unitcell = crystal(symbols=['H', 'O', 'X', 'Fe'],
                           basis=[(0., 0., 0.), (0.39, 0.14, 0.),
                                  (0.2, 0.35, 0.5), (0.22, 0.38, 0.)],
                           spacegroup=55,
                           cellpar=[6.25, 7.4, 3.83, 90, 90, 90],
                           size=[1, 1, 1])

        for atom in unitcell:
            atom.tag = atom.index

        atoms = wrap_and_sort_by_position(unitcell*(3, 3, 2))

        symm_group = [atom.tag for atom in atoms]
        index_by_group = [[] for _ in range(len(unitcell))]

        for atom in atoms:
            index_by_group[atom.tag].append(atom.index)
        ref_indx = [min(gr) for gr in index_by_group]

        tm_brute = brute_force_tm_construction(ref_indx, index_by_group, atoms)

        tm_constructor = TransMatrixConstructor(atoms, 5.0)
        tm_fast = tm_constructor.construct(ref_indx, symm_group)
        self.assertTrue(check_sparse_dense(tm_fast, tm_brute))

    def test_sp217(self):
        a = 10.553
        b = 10.553
        c = 10.553
        alpha = 90
        beta = 90
        gamma = 90
        cellpar = [a, b, c, alpha, beta, gamma]
        basis = [(0, 0, 0), (0.324, 0.324, 0.324),
                 (0.3582, 0.3582, 0.0393), (0.0954, 0.0954, 0.2725)]

        unitcell = crystal(symbols=['Al', 'Al', 'Al', 'Al'], cellpar=cellpar,
                           spacegroup=217, primitive_cell=False,
                           basis=basis)

        # Tag atoms
        for atom in unitcell:
            atom.tag = atom.index

        atoms = wrap_and_sort_by_position(unitcell*(2, 2, 2))
        symm_group = [atom.tag for atom in atoms]

        index_by_group = [[] for _ in range(len(unitcell))]

        for atom in atoms:
            index_by_group[atom.tag].append(atom.index)
        ref_indx = [min(gr) for gr in index_by_group]

        tm_brute = brute_force_tm_construction(ref_indx, index_by_group, atoms)

        tm_constructor = TransMatrixConstructor(atoms, 5.0)
        tm_fast = tm_constructor.construct(ref_indx, symm_group)
        self.assertTrue(check_sparse_dense(tm_fast, tm_brute))


def timing():
    import time
    import numpy as np
    atoms = bulk("Al")

    brute_force_time = []
    fast_time = []
    sizes = []
    for n in range(2, 15):
        print("Current size: {}".format(n))
        atoms_scaled = atoms*(n, n, n)
        atoms_scaled = wrap_and_sort_by_position(atoms_scaled)

        symm_group = [0 for _ in range(len(atoms_scaled))]
        ref_indx = [0]
        index_by_group = [list(range(len(atoms_scaled)))]

        start = time.time()
        brute_force_tm_construction(ref_indx, index_by_group, atoms_scaled)
        brute_force_time.append(time.time() - start)

        start = time.time()
        fast_tm = TransMatrixConstructor(atoms_scaled, 5.0)
        fast_tm.construct(ref_indx, symm_group)
        fast_time.append(time.time() - start)

        sizes.append(n)

    np.savetxt("trans_mat_timing.csv",
               np.vstack((sizes, brute_force_time, fast_time)).T,
               header="Size, Brute force, fast")

# timing()


if __name__ == '__main__':
    unittest.main()



from clease.gui.cleaseGUI import CleaseGUI
import unittest
import time
from kivy.clock import Clock
from functools import partial
from settings_page_tests import SettingsPageTests
from concentraton_page_tests import ConcentrationPageTest
from new_struct_page_test import NewStructPageTest
from test_fit_page import FitPageTests
from actionbar_tests import ActionBarTests


class TestSettingsPage(unittest.TestCase):
    framecount = 0
    interval = 0.0001

    def pause(*args):
        time.sleep(self.interval)

    def run_tests(self, app):
        """
        Main function for running all tests. Note that the application
        is never restarted. Therefore, actions follow each other
        """
        Clock.schedule_interval(self.pause, self.interval)

        input_page_tests = SettingsPageTests()
        input_page_tests.run_with_app(app)

        conc_page_test = ConcentrationPageTest()
        conc_page_test.run_with_app(app)

        new_struct_test = NewStructPageTest()
        new_struct_test.run_with_app(app)

        fit_page_test = FitPageTests()
        fit_page_test.run_with_app(app)

        actionbar_test = ActionBarTests()
        actionbar_test.run_with_app(app)

        app.stop()

    def test_gui(self):
        app = CleaseGUI()
        Clock.schedule_once(lambda x: self.run_tests(app),
                            self.interval)
        app.run()

if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch
from clease.gui.newStructPage import NewStructPage
from kivy.lang import Builder
from kivy.resources import resource_add_path
import os.path as op
import clease.gui
import json
import os
from unittest.mock import patch, MagicMock, PropertyMock
from clease.gui.newStructPage import NewStructPage
from clease import CEBulk
from ase.build import bulk
from ase.io.trajectory import TrajectoryWriter
from ase.io import write


class NewStructPageTest(unittest.TestCase):
    def load_pop_ups(self, app):
        screen = app.root.ids.sm.get_screen('NewStruct')

        self.assertTrue(screen._pop_up is None)
        screen.ids.loadInitStruct.dispatch('on_release')
        self.assertFalse(screen._pop_up is None)
        self.assertEqual(screen._pop_up.title, "Load initial structure")
        screen.dismiss_popup()

        self.assertTrue(screen._pop_up is None)
        screen.ids.loadFinalStruct.dispatch('on_release')
        self.assertFalse(screen._pop_up is None)
        self.assertEqual(screen._pop_up.title, "Load final structure")
        screen.dismiss_popup()

        self.assertTrue(screen._pop_up is None)
        screen.ids.loadECIFile.dispatch('on_release')
        self.assertFalse(screen._pop_up is None)
        self.assertEqual(screen._pop_up.title, "Load ECI filename")
        screen.dismiss_popup()

        self.assertTrue(screen._pop_up is None)
        screen.ids.loadTemplateAtoms.dispatch('on_release')
        self.assertFalse(screen._pop_up is None)
        self.assertEqual(screen._pop_up.title, "Load template atoms")
        screen.dismiss_popup()

    @patch('clease.NewStructures')
    @patch('clease.gui.newStructPage.App')
    def test_exhautive_gui(self, app_mock, new_struct_mock):
        page = NewStructPage()

        # Change to Exhaustive-search
        page.ids.newStructTypeSpinner.text = 'Ground-state structure (variable template)'

        # Check active fields
        ids = page.ids
        self.assertFalse(ids.tempMaxInput.disabled)
        self.assertFalse(ids.tempMinInput.disabled)
        self.assertFalse(ids.numSweepsInput.disabled)
        self.assertFalse(ids.eciFileInput.disabled)
        self.assertFalse(ids.loadECIFile.disabled)
        self.assertFalse(ids.numTemplateInput.disabled)
        self.assertFalse(ids.numPrimCellsInput.disabled)
        self.assertTrue(ids.loadTemplateAtoms.disabled)
        self.assertTrue(ids.templateAtomsInput.disabled)
        self.assertTrue(ids.randomizeCompositionSpinner.disabled)

        # Create an ECI file
        ecis = {'c0': 1.0}
        eci_file = 'example_eci_new_struct_page.json'
        with open(eci_file, 'w') as out:
            json.dump(ecis, out)

        # Update ECI file
        page.ids.eciFileInput.text = eci_file

        # Try to generate a structure
        page.ids.generateButton.dispatch('on_release')

        # Make sure functions where called correctly
        method = new_struct_mock().generate_gs_structure_multiple_templates
        os.remove(eci_file)
        self.assertEqual(method.call_count, 1)
 
    @patch('clease.gui.newStructPage.App')
    @patch('clease.NewStructures')
    def test_import_structures(self, new_struct_mock, kivy_mock):
        init_file = 'initial_structure.traj'
        final_file = 'final_structure.traj'
        traj_init = TrajectoryWriter(init_file)
        traj_final = TrajectoryWriter(final_file)

        for _ in range(2):
            atoms = bulk('Au')
            traj_init.write(atoms)
            traj_final.write(atoms)

        page = NewStructPage()
        page.ids.initStructInput = MagicMock(text=init_file)
        page.ids.finalStructInput = MagicMock(text=final_file)
        page.import_structures()

        # Make sure the method is called exactly once
        method = new_struct_mock().insert_structures
        self.assertEqual(method.call_count, 1)

        # Make sure that insert_structures was called with the right arguments
        method.assert_called_with(traj_init=init_file, traj_final=final_file)
        del traj_init
        del traj_final
        os.remove(init_file)
        os.remove(final_file)

        # Try to read xyz files
        init_file = 'initial_structure.xyz'
        final_file = 'final_structure.xyz'
        atoms = bulk('Au')
        write(init_file, atoms)
        write(final_file, atoms)
        page.ids.initStructInput = MagicMock(text=init_file)
        page.ids.finalStructInput = MagicMock(text=final_file)
        page.import_structures()

        method = new_struct_mock().insert_structure
        self.assertEqual(method.call_count, 1)
        os.remove(init_file)
        os.remove(final_file)

    def run_with_app(self, app):
        self.load_pop_ups(app)

if __name__ == '__main__':
    # Load the layout for the new struct page
    main_path = op.abspath(clease.gui.__file__)
    main_path = main_path.rpartition("/")[0]
    resource_add_path(main_path + '/layout')
    Builder.load_file("newStructPageLayout.kv")
    unittest.main()



from clease.tools import dec_string, list2str
from itertools import product
from clease.tools import flatten


class ClusterList(object):
    def __init__(self):
        self.clusters = []

    def append(self, cluster):
        self.clusters.append(cluster)

    def clear(self):
        """Clear the content."""
        self.clusters = []

    @property
    def names(self):
        return list(set([c.name for c in self.clusters]))

    def get_by_name(self, name):
        return [c for c in self.clusters if c.name == name]

    def get_by_name_and_group(self, name, group):
        clusters = self.get_by_name(name)
        for c in clusters:
            if c.group == group:
                return c
        raise ValueError("There are no cluster named {} and group {}"
                         "".format(name, group))

    def get_by_size(self, size):
        # Return all clusters with a given size
        return [c for c in self.clusters if c.size == size]

    def get_by_group(self, group):
        """Return all clusters in a given symmetry group."""
        return [c for c in self.clusters if c.group == group]

    def get_equivalent_clusters(self, cluster):
        """Return equivalent clusters in other symmetry groups."""
        equiv = []
        for c in self.clusters:
            if c == cluster:
                equiv.append(c)
        return equiv

    @staticmethod
    def get_cf_names(cluster, num_bf):
        """Return all possible correlation function names.

        Parameters:

        cluster: Cluster
            Instance of cluster class

        num_bf: int
            Number of basis functions
        """
        name = cluster.name
        eq_sites = cluster.equiv_sites
        bf_list = list(range(num_bf))
        cf_names = []
        for dec in product(bf_list, repeat=cluster.size):
            dec_str = dec_string(dec, eq_sites)
            cf_names.append(name + '_' + dec_str)
        return sorted(list(set(cf_names)))

    def get_all_cf_names(self, num_bf):
        """
        Return a list of all correlation function names

        Parameters:

        num_bf: int
            Number of basis functions
        """
        all_cf_names = []
        for cluster in self.clusters:
            if cluster.name == 'c0':
                all_cf_names.append('c0')
            else:
                all_cf_names += self.get_cf_names(cluster, num_bf)
        return sorted(list(set(all_cf_names)))

    def __len__(self):
        return len(self.clusters)

    def __getitem__(self, index):
        return self.clusters[index]

    def __array__(self):
        return self.clusters

    def __eq__(self, other):
        if len(self.clusters) != len(other.clusters):
            return False

        for c1, c2 in zip(self.clusters, other.clusters):
            if c1 != c2:
                print(c1, c2)
                return False
        return True

    def sort(self):
        self.clusters.sort()

    @property
    def dtype(self):
        return 'ClusterList'

    def tolist(self):
        return self.clusters

    @property
    def unique_indices(self):
        all_indices = set()
        for cluster in self.clusters:
            all_indices.update(flatten(cluster.indices))
        return list(all_indices)

    @property
    def num_symm_groups(self):
        return max(c.group for c in self.clusters)

    @property
    def unique_indices_per_group(self):
        indices_per_group = []
        for i in range(self.num_symm_groups + 1):
            indices_per_group.append(set())
            for c in self.get_by_group(i):
                indices_per_group[i].update(flatten(c.indices))
        return [list(x) for x in indices_per_group]

    def multiplicity_factors(self, num_sites_per_group):
        mult_factors = {}
        norm = {}
        for cluster in self.clusters:
            current_factor = mult_factors.get(cluster.name, 0)
            current_norm = norm.get(cluster.name, 0)
            current_factor += \
                len(cluster.indices)*num_sites_per_group[cluster.group]
            current_norm += num_sites_per_group[cluster.group]

            mult_factors[cluster.name] = current_factor
            norm[cluster.name] = current_norm

        for k in mult_factors.keys():
            mult_factors[k] /= norm[k]
        return mult_factors

    def get_subclusters(self, cluster):
        """Return all all subclusters of the passed cluster in the list."""
        return [c for c in self.clusters if c.is_subcluster(cluster)]

    def get_figures(self, atoms):
        figures = []
        self.clusters.sort()
        for cluster in self.clusters:
            if cluster.name == 'c0' or cluster.name == 'c1':
                continue
            figure = cluster.get_figure(atoms)
            figure.info = {'name': cluster.name}
            figures.append(figure)
        return figures

    def get_occurence_counts(self):
        """Count the number of occurences of all figures."""
        occ_counts = []
        for cluster in self.clusters:
            occ_counts.append(cluster.num_fig_occurence)
        return occ_counts

    def num_occ_figure(self, fig_key, c_name, symm_groups, trans_matrix):
        """Determine the number of occurences of a figure across the structure.

        Parameter:

        fig_key: str
            string containing the indices of consituting atomic indices

        c_name: str
            name of the cluster of the figure

        symm_groups: list
            list containing the symmetry group number of the atomic indices

        tran_matrix: list of dicts
            translation matrix
        """
        figure = list(map(int, fig_key.split("-")))
        tot_num = 0
        clusters = self.get_by_name(c_name)
        occ_count = {}
        for cluster in clusters:
            occ_count[cluster.group] = cluster.num_fig_occurences

        for ref in set(figure):
            group = symm_groups[ref]
            cluster = self.get_by_name_and_group(c_name, group)
            corr_figure = \
                cluster.corresponding_figure(ref, figure, trans_matrix)
            corr_fig_key = list2str(corr_figure)
            tot_num += occ_count[cluster.group][corr_fig_key]
        return tot_num


"""Test to initiatialize CE using a CECrystal."""
import os
import json
from clease import CECrystal, NewStructures, CorrFunction
from clease.newStruct import MaxAttemptReachedError
from clease.concentration import Concentration
from clease.tools import wrap_and_sort_by_position
from ase.db import connect
from reference_corr_funcs_crystal import all_cf
from ase.spacegroup import crystal
import unittest

# If this is True, the JSON file containing the correlation functions
# Used to check consistency of the reference functions is updated
# This should normally be False
update_reference_file = False

tol = 1E-9


def get_figures_of_family(setting, cname):
    """Return the figures of a given cluster family."""
    figures = []
    clusters = setting.cluster_list.get_by_name(cname)
    for cluster in clusters:
        figures.append(cluster.indices)
    return figures


class TestCECrystal(unittest.TestCase):
    def test_spgroup_217(self):
        """Test the initialization of spacegroup 217."""
        db_name = "test_spacegroup_217.db"
        a = 10.553
        b = 10.553
        c = 10.553
        alpha = 90
        beta = 90
        gamma = 90
        cellpar = [a, b, c, alpha, beta, gamma]
        basis = [(0, 0, 0), (0.324, 0.324, 0.324),
                 (0.3582, 0.3582, 0.0393), (0.0954, 0.0954, 0.2725)]
        basis_elements = [["Al", "Mg"], ["Al", "Mg"], ["Al", "Mg"],
                          ["Al", "Mg"]]

        # Test with grouped basis
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=[[0, 1, 2, 3]])
        bsg = CECrystal(basis=basis,
                        spacegroup=217,
                        cellpar=cellpar,
                        concentration=concentration,
                        max_cluster_size=3,
                        db_name=db_name,
                        size=[1, 1, 1],
                        max_cluster_dia=[3.5, 3.5])

        # The correlation functions are actually calculated for the
        # conventional cell
        atoms = crystal(symbols=['Al', 'Al', 'Al', 'Al'], cellpar=cellpar,
                        spacegroup=217, primitive_cell=False,
                        basis=basis)
        atoms = wrap_and_sort_by_position(atoms)
        self.assertEqual(bsg.num_trans_symm, 29)
        # atoms = bsg.atoms.copy()
        atoms[0].symbol = "Mg"
        atoms[10].symbol = "Mg"
        atoms[20].symbol = "Mg"
        atoms[30].symbol = "Mg"
        corr = CorrFunction(bsg)
        cf = corr.get_cf(atoms)

        if update_reference_file:
            all_cf["sp_217_grouped"] = cf
        for key in cf.keys():
            self.assertAlmostEqual(cf[key], all_cf["sp_217_grouped"][key])

        os.remove(db_name)

    def test_two_grouped_basis(self):
        db_name = "test_spacegroup_two_grouped.db"
        # ---------------------------------- #
        # 2 grouped_basis                    #
        # ---------------------------------- #
        basis_elements = [['Li', 'X', 'V'], ['Li', 'X', 'V'],
                          ['O', 'F']]
        grouped_basis = [[0, 1], [2]]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=grouped_basis)

        cellpar = [5.123, 5.123, 13.005, 90., 90., 120.]
        basis = [(0.00, 0.00, 0.00),
                 (1. / 3, 2. / 3, 0.00),
                 (1. / 3, 0.00, 0.25)]
        bsg = CECrystal(basis=basis,
                        spacegroup=167,
                        cellpar=cellpar,
                        concentration=concentration,
                        size=[1, 1, 1],
                        db_name=db_name,
                        max_cluster_size=3,
                        max_cluster_dia=[2.5, 2.5])
        self.assertTrue(bsg.unique_elements == ['F', 'Li', 'O', 'V', 'X'])
        self.assertTrue(bsg.spin_dict == {'F': 2.0, 'Li': -2.0,
                                          'O': 1.0, 'V': -1.0, 'X': 0})
        self.assertEqual(len(bsg.basis_functions), 4)
        self.assertEqual(bsg.num_basis, 2)
        self.assertEqual(len(bsg.index_by_basis), 2)
        self.assertEqual(len(bsg.basis_functions), 4)

        # The correlation functions in the reference file is calculated for
        # the conventional cell
        atoms = crystal(symbols=['Li', 'Li', 'O'], basis=basis,
                        cellpar=cellpar, primitive_cell=False,
                        spacegroup=167)
        atoms = wrap_and_sort_by_position(atoms)
        indx_to_X = [6, 33, 8, 35]
        for indx in indx_to_X:
            atoms[indx].symbol = "X"
        corr = CorrFunction(bsg)
        cf = corr.get_cf(atoms)
        if update_reference_file:
            all_cf["Li_X_V_O_F"] = cf
        for key in cf.keys():
            self.assertAlmostEqual(cf[key], all_cf["Li_X_V_O_F"][key])

        os.remove(db_name)

    def test_two_grouped_basis_probe_structure(self):
        # ---------------------------------- #
        # 2 grouped_basis                    #
        # ---------------------------------- #
        # initial_pool + probe_structures    #
        # ---------------------------------- #
        db_name = "test_grouped_probe.db"
        basis_elements = [['O', 'X'], ['O', 'X'],
                          ['O', 'X'], ['Ta']]
        grouped_basis = [[0, 1, 2], [3]]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=grouped_basis)

        bsg = CECrystal(basis=[(0., 0., 0.),
                               (0.3894, 0.1405, 0.),
                               (0.201, 0.3461, 0.5),
                               (0.2244, 0.3821, 0.)],
                        spacegroup=55,
                        cellpar=[6.25, 7.4, 3.83, 90, 90, 90],
                        size=[1, 2, 2],
                        concentration=concentration,
                        db_name=db_name,
                        max_cluster_size=3,
                        max_cluster_dia=[3.0, 3.0])

        self.assertTrue(bsg.unique_elements == ['O', 'Ta', 'X'])
        self.assertTrue(bsg.spin_dict == {'O': 1.0, 'Ta': -1.0, 'X': 0.0})
        self.assertEqual(len(bsg.basis_functions), 2)
        self.assertEqual(bsg.num_basis, 2)
        self.assertEqual(len(bsg.index_by_basis), 2)
        self.assertEqual(len(bsg.basis_functions), 2)

        atoms = bsg.atoms.copy()
        indx_to_X = [0, 4, 8, 12, 16]
        for indx in indx_to_X:
            atoms[indx].symbol = "X"
        corr = CorrFunction(bsg)
        cf = corr.get_cf(atoms)
        if update_reference_file:
            all_cf["Ta_O_X_grouped"] = cf
        for key in cf.keys():
            self.assertAlmostEqual(cf[key], all_cf["Ta_O_X_grouped"][key])

        try:
            ns = NewStructures(setting=bsg, struct_per_gen=2)
            ns.generate_initial_pool()
            ns = NewStructures(setting=bsg, struct_per_gen=2)
            ns.generate_probe_structure(init_temp=1.0, final_temp=0.001,
                                        num_temp=5, num_steps_per_temp=100,
                                        approx_mean_var=True)

            db = connect(db_name)
            for row in db.select(struct_type='initial'):
                atoms = row.toatoms()
                cf = corr.get_cf(atoms)
                for key, value in cf.items():
                    self.assertAlmostEqual(row["polynomial_cf"][key], value)

        except MaxAttemptReachedError as exc:
            print(str(exc))

        os.remove(db_name)

    def test_two_grouped_basis_background_atoms_probe_structure(self):
        # ---------------------------------- #
        # 2 grouped_basis + background atoms #
        # ---------------------------------- #
        # initial_pool + probe_structures    #
        # ---------------------------------- #
        db_name = "test_grouped_probe_back_probe.db"
        basis_elements = [['O', 'X'], ['Ta'], ['O', 'X'], ['O', 'X']]
        grouped_basis = [[1], [0, 2, 3]]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=grouped_basis)

        bsg = CECrystal(basis=[(0., 0., 0.),
                               (0.2244, 0.3821, 0.),
                               (0.3894, 0.1405, 0.),
                               (0.201, 0.3461, 0.5)],
                        spacegroup=55,
                        cellpar=[6.25, 7.4, 3.83, 90, 90, 90],
                        size=[2, 2, 3],
                        concentration=concentration,
                        db_name=db_name,
                        max_cluster_size=3,
                        max_cluster_dia=[3.0, 3.0],
                        ignore_background_atoms=True)

        self.assertTrue(bsg.unique_elements == ['O', 'Ta', 'X'])
        self.assertTrue(bsg.spin_dict == {'O': 1.0, 'X': -1.0})
        self.assertTrue(bsg.basis_elements == [['Ta'], ['O', 'X']])
        self.assertEqual(len(bsg.basis_functions), 1)
        self.assertEqual(bsg.num_basis, 2)
        self.assertEqual(len(bsg.index_by_basis), 2)
        self.assertEqual(len(bsg.basis_functions), 1)

        try:
            ns = NewStructures(setting=bsg, struct_per_gen=2)
            ns.generate_initial_pool()
            ns = NewStructures(setting=bsg, struct_per_gen=2)
            ns.generate_probe_structure(init_temp=1.0, final_temp=0.001,
                                        num_temp=5, num_steps_per_temp=100,
                                        approx_mean_var=True)
            atoms = bsg.atoms.copy()
            indx_to_X = [0, 4, 8, 12, 16]
            for indx in indx_to_X:
                atoms[indx].symbol = "X"
            corr = CorrFunction(bsg)
            cf = corr.get_cf(atoms)
            if update_reference_file:
                all_cf["Ta_O_X_ungrouped"] = cf

            for key in cf.keys():
                self.assertAlmostEqual(cf[key],
                                       all_cf["Ta_O_X_ungrouped"][key])

            db = connect(db_name)
            for row in db.select(struct_type='initial'):
                atoms = row.toatoms()
                cf = corr.get_cf(atoms)
                for key, value in cf.items():
                    self.assertAlmostEqual(row["polynomial_cf"][key], value)

        except MaxAttemptReachedError as exc:
            print(str(exc))

        os.remove(db_name)

    def test_narrow_angle_crystal(self):
        """Test that Probestructure works for crystals with narrow angles.

        This test a crystal with internal angles 50, 20, 15 degree.
        """
        db_name = "test_grouped_narrow_angle.db"
        basis_elements = [['Mg', 'Si']]
        concentration = Concentration(basis_elements=basis_elements)
        bsg = CECrystal(basis=[(0.0, 0.0, 0.0)],
                        spacegroup=225,
                        cellpar=[4.0, 4.0, 4.0, 50.0, 40.0, 15.0],
                        concentration=concentration,
                        db_name=db_name,
                        size=[2, 2, 1],
                        max_cluster_size=3,
                        max_cluster_dia=[1.05, 1.05])

        assert len(bsg.index_by_trans_symm) == 1

        try:
            ns = NewStructures(setting=bsg, struct_per_gen=2)
            ns.generate_initial_pool()
            ns = NewStructures(setting=bsg, struct_per_gen=2)
            ns.generate_probe_structure(init_temp=1.0, final_temp=0.001,
                                        num_temp=5, num_steps_per_temp=100,
                                        approx_mean_var=True)
        except MaxAttemptReachedError as exc:
            print(str(exc))
        os.remove(db_name)

    def tearDown(self):
        if update_reference_file:
            print("Updating the reference correlation function file")
            print("This should normally not be done.")
            with open("reference_corr_funcs_crystal.py", 'w') as outfile:
                json.dump(all_cf, outfile, indent=2, separators=(',', ': '))
        return super().tearDown()


if __name__ == '__main__':
    unittest.main()





from clease import ClusterInfoMapper, CEBulk, Concentration, CorrFunction
from clease import CECrystal
from clease.tools import wrap_and_sort_by_position
from clease.cluster_info_mapper import AtomsNotContainedInLargeCellError
import unittest
import os
from ase.build import bulk
from ase.spacegroup import crystal


def dict_amost_equal(d1, d2):
    for k, v in d1.items():
        if abs(v - d2[k]) > 1E-6:
            return False
    return True


class TestClusterInfoMapper(unittest.TestCase):
    def test_fcc(self):
        db_name = "test_bulk_info_map_fcc.db"
        basis_elements = [['Au', 'Cu']]
        concentration = Concentration(basis_elements=basis_elements)
        setting = CEBulk(crystalstructure='fcc', a=4.05, size=[6, 6, 6],
                         concentration=concentration, db_name=db_name,
                         max_cluster_dia=[4.3, 4.3, 4.3],
                         max_cluster_size=4)

        # Let's try to obtain cluster info for conventional cell
        atoms_small = bulk("Au", crystalstructure="fcc", cubic=True, a=4.05)
        atoms_small = atoms_small*(2, 1, 1)
        atoms_small = wrap_and_sort_by_position(atoms_small)

        info_mapper = ClusterInfoMapper(
            setting.atoms, setting.trans_matrix, setting.cluster_list)

        map_info, map_tm = info_mapper.map_info(atoms_small)

        atoms_small[0].symbol = 'Cu'
        atoms_small[4].symbol = 'Cu'

        # Generate the cubic from scratch
        setting.set_active_template(atoms=atoms_small, generate_template=True)

        cf = CorrFunction(setting)
        cf1 = cf.get_cf(atoms_small)

        # Change the info to the mapped one
        setting.cluster_list = map_info
        setting.trans_matrix = map_tm
        cf2 = cf.get_cf(atoms_small)

        os.remove(db_name)
        self.assertTrue(dict_amost_equal(cf1, cf2))

    def test_TaXO(self):
        db_name = 'cluster_info_mapper_TaXO.db'
        basis_elements = [['O', 'X'], ['Ta'], ['O', 'X'], ['O', 'X']]
        grouped_basis = [[1], [0, 2, 3]]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=grouped_basis)

        bsg = CECrystal(basis=[(0., 0., 0.),
                               (0.2244, 0.3821, 0.),
                               (0.3894, 0.1405, 0.),
                               (0.201, 0.3461, 0.5)],
                        spacegroup=55,
                        cellpar=[6.25, 7.4, 3.83, 90, 90, 90],
                        size=[4, 4, 4],
                        concentration=concentration,
                        db_name=db_name,
                        max_cluster_size=3,
                        max_cluster_dia=[3.0, 3.0],
                        ignore_background_atoms=True)

        # Create a small cell
        atoms_small = crystal(basis=[(0., 0., 0.),
                                     (0.2244, 0.3821, 0.),
                                     (0.3894, 0.1405, 0.),
                                     (0.201, 0.3461, 0.5)],
                              spacegroup=55,
                              cellpar=[6.25, 7.4, 3.83, 90, 90, 90],
                              symbols=['O', 'Ta', 'O', 'O'])
        atoms_small = atoms_small*(2, 1, 1)

        atoms_small = atoms_small = wrap_and_sort_by_position(atoms_small)

        info_mapper = \
            ClusterInfoMapper(bsg.atoms, bsg.trans_matrix, bsg.cluster_list)

        map_info, map_tm = info_mapper.map_info(atoms_small)

        # Swap 3 O with X
        count = 0
        for atom in atoms_small:
            if atom.symbol == 'O':
                atom.symbol = 'X'
                count += 1

            if count >= 3:
                break

        # Generate the cubic from scratch
        bsg.set_active_template(atoms=atoms_small, generate_template=True)

        cf = CorrFunction(bsg)
        cf1 = cf.get_cf(atoms_small)

        # Change the info to the mapped one
        bsg.cluster_list = map_info
        bsg.trans_matrix = map_tm
        cf2 = cf.get_cf(atoms_small)

        os.remove(db_name)
        self.assertTrue(dict_amost_equal(cf1, cf2))

    def test_not_contained_error(self):
        atoms = bulk("Al", crystalstructure="fcc", a=4.05)*(2, 1, 1)
        atoms2 = bulk("Al", crystalstructure="fcc", a=4.05)*(1, 1, 3)

        info_mapper = ClusterInfoMapper(atoms, None, None)
        with self.assertRaises(AtomsNotContainedInLargeCellError):
            info_mapper._map_indices(atoms2)


if __name__ == '__main__':
    unittest.main()



"""Test suite for TemplateAtoms."""
import os
from clease.template_atoms import TemplateAtoms
from ase.build import bulk
from ase.db import connect
from ase.build import niggli_reduce
from clease import Concentration, ValidConcentrationFilter
from clease.template_filters import AtomsFilter, CellFilter
from clease import DistanceBetweenFacetsFilter
from clease import Concentration, CEBulk
import numpy as np
import unittest
from unittest.mock import patch


class SettingsPlaceHolder(object):
    """
    Dummy object that simply holds the few variables needed for the test.
    Only purpose of this is to make the test fast
    """
    atoms = None
    index_by_basis = []
    Concentration = None


class NumAtomsFilter(AtomsFilter):
    def __init__(self, min_num_atoms):
        self.min_num_atoms = min_num_atoms

    def __call__(self, atoms):
        return len(atoms) > self.min_num_atoms


class DummyCellFilter(CellFilter):
    def __call__(self, cell):
        return True


class TestTemplates(unittest.TestCase):
    def test_fcc(self):
        db_name = 'templates_fcc.db'
        prim_cell = bulk("Cu", a=4.05, crystalstructure='fcc')
        db = connect(db_name)
        db.write(prim_cell, name='primitive_cell')

        template_atoms = TemplateAtoms(supercell_factor=27, size=None,
                                       skew_threshold=4,
                                       db_name=db_name)
        dims = template_atoms.get_size()
        ref = [[1, 1, 1], [1, 1, 2], [2, 2, 2], [2, 2, 3], [2, 2, 4],
               [2, 2, 5], [2, 3, 3], [2, 3, 4], [3, 3, 3]]

        ref = [np.diag(x).tolist() for x in ref]
        self.assertTrue(dims, ref)

        os.remove(db_name)

    def test_hcp(self):
        db_name = 'templates_hcp.db'
        prim_cell = bulk("Mg")
        db = connect(db_name)
        db.write(prim_cell, name='primitive_cell')
        template_atoms = TemplateAtoms(supercell_factor=27, size=None,
                                       skew_threshold=5, db_name=db_name)
        dims = template_atoms.get_size()
        ref = [[1, 1, 1], [1, 1, 2], [1, 2, 1], [1, 2, 2], [1, 3, 1],
               [1, 3, 2], [1, 4, 1], [2, 2, 1], [2, 2, 2], [2, 2, 3],
               [2, 2, 4], [2, 2, 5], [2, 3, 1], [2, 3, 2], [2, 3, 3],
               [2, 3, 4], [2, 4, 1], [2, 4, 2], [2, 4, 3], [2, 5, 1],
               [2, 5, 2], [2, 6, 1], [2, 6, 2], [3, 3, 1], [3, 3, 2],
               [3, 3, 3], [3, 4, 1], [3, 4, 2], [3, 5, 1], [3, 6, 1],
               [4, 4, 1], [4, 5, 1]]

        ref = [np.diag(x).tolist() for x in ref]
        self.assertTrue(dims, ref)

        os.remove(db_name)

    def test_valid_concentration_filter(self):
        prim_cell = bulk("NaCl", crystalstructure="rocksalt", a=4.0)
        settings = SettingsPlaceHolder()
        settings.atoms = prim_cell
        settings.index_by_basis = [[0], [1]]

        db_name = 'test_valid_concentration.db'
        db = connect(db_name)
        db.write(prim_cell, name='primitive_cell')

        # Force vacancy concentration to be exactly 2/3 of the Cl
        # concentration
        A_eq = [[0, 1, -2.0]]
        b_eq = [0.0]
        settings.concentration = Concentration(
            basis_elements=[['Na'], ['Cl', 'X']], A_eq=A_eq, b_eq=b_eq)

        template_generator = TemplateAtoms(
            db_name=db_name, supercell_factor=20,
            skew_threshold=1000000000)

        os.remove(db_name)
        conc_filter = ValidConcentrationFilter(settings)
        # Check that you cannot attach an AtomsFilter as a cell
        # filter
        with self.assertRaises(TypeError):
            template_generator.add_cell_filter(conc_filter)

        template_generator.clear_filters()
        template_generator.add_atoms_filter(conc_filter)

        templates = template_generator._generate_template_atoms()

        for atoms in templates['atoms']:
            num_cl = sum(1 for atom in atoms if atom.symbol == 'Cl')
            self.assertAlmostEqual(2.0*num_cl/3.0, np.round(2.0*num_cl/3.0))

    def test_dist_filter(self):
        f = DistanceBetweenFacetsFilter(4.0)
        cell = [[0.1, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0]]
        cell = np.array(cell)
        self.assertFalse(f(cell))
        cell[0, 0] = 0.3
        self.assertTrue(f(cell))

    def test_fixed_vol(self):
        db_name = 'templates_fcc_fixed_vol.db'
        prim_cell = bulk("Cu", a=4.05, crystalstructure='fcc')
        db = connect(db_name)
        db.write(prim_cell, name='primitive_cell')

        template_atoms = TemplateAtoms(supercell_factor=27, size=None,
                                       skew_threshold=4,
                                       db_name=db_name)
        templates = template_atoms.get_fixed_volume_templates(
            num_prim_cells=4, num_templates=100)
        os.remove(db_name)

        # Conform that the conventional cell is present
        found_conventional = False
        conventional = [4.05, 4.05, 4.05, 90, 90, 90]
        for atoms in templates:
            niggli_reduce(atoms)
            lengths_ang = atoms.get_cell_lengths_and_angles()
            if np.allclose(lengths_ang, conventional):
                found_conventional = True
                break
        self.assertTrue(found_conventional)

    @patch('test_templates.CEBulk._read_data')
    @patch('test_templates.CEBulk._store_data')
    @patch('test_templates.CEBulk.create_cluster_list_and_trans_matrix')
    def test_fixed_vol_with_conc_constraint(self, *args):
        A_eq = [[3, -2]]
        b_eq = [0]
        conc = Concentration(basis_elements=[['Au', 'Cu']],
                             A_eq=A_eq, b_eq=b_eq)

        db_name = 'test_fixed_vol_conc_constraint.db'
        setting = CEBulk(crystalstructure='fcc', a=3.8, size=[1, 1, 5],
                         db_name=db_name, max_cluster_size=2,
                         max_cluster_dia=3.0, concentration=conc)

        tmp = setting.template_atoms
        v_conc = ValidConcentrationFilter(setting)
        tmp.add_atoms_filter(v_conc)

        sizes = [4, 5, 7, 10]
        valid_size = [5, 10]
        for s in sizes:
            templates = tmp.get_fixed_volume_templates(num_prim_cells=s)

            if s in valid_size:
                self.assertGreater(len(templates), 0)
            else:
                self.assertEqual(len(templates), 0)
        os.remove(db_name)

    def test_apply_filter(self):
        db_name = 'templates_fcc_apply_filter.db'
        prim_cell = bulk("Cu", a=4.05, crystalstructure='fcc')
        db = connect(db_name)
        db.write(prim_cell, name='primitive_cell')

        template_atoms = TemplateAtoms(supercell_factor=27, size=None,
                                       skew_threshold=4,
                                       db_name=db_name)

        num_atoms = 16
        # First confirm that we have cells with less than 16 atoms
        has_less_than = False
        for atoms in template_atoms.templates['atoms']:
            if len(atoms) < num_atoms:
                has_less_than = True
                break
        self.assertTrue(has_less_than)

        # Filter all atoms with less than 16 atoms
        template_atoms.apply_filter(NumAtomsFilter(num_atoms))
        for atoms in template_atoms.templates['atoms']:
            self.assertGreaterEqual(len(atoms), num_atoms)
        os.remove(db_name)

    def test_remove_atoms_filter(self):
        db_name = 'templates_remove_atoms_filter.db'
        prim_cell = bulk("Cu", a=4.05, crystalstructure='fcc')
        db = connect(db_name)
        db.write(prim_cell, name='primitive_cell')

        template_atoms = TemplateAtoms(supercell_factor=3, size=None,
                                       skew_threshold=4,
                                       db_name=db_name)

        f = NumAtomsFilter(16)
        template_atoms.add_atoms_filter(f)
        self.assertEqual(len(template_atoms.atoms_filters), 1)
        template_atoms.remove_filter(f)
        self.assertEqual(len(template_atoms.atoms_filters), 0)
        os.remove(db_name)

    def test_remove_cell_filter(self):
        db_name = 'templates_remove_atoms_filter.db'
        prim_cell = bulk("Cu", a=4.05, crystalstructure='fcc')
        db = connect(db_name)
        db.write(prim_cell, name='primitive_cell')

        template_atoms = TemplateAtoms(supercell_factor=3, size=None,
                                       skew_threshold=4,
                                       db_name=db_name)

        num_cell_filters = len(template_atoms.cell_filters)
        f = DummyCellFilter()
        template_atoms.add_cell_filter(f)
        self.assertEqual(len(template_atoms.cell_filters), num_cell_filters+1)
        template_atoms.remove_filter(f)
        self.assertEqual(len(template_atoms.cell_filters), num_cell_filters)
        os.remove(db_name)


if __name__ == '__main__':
    unittest.main()



"""Test case for the multiprocessing logger."""
import multiprocessing as mp
from clease.mp_logger import MultiprocessHandler
import logging as lg
import os
import unittest

if __name__ != "__main__":
    msg = "Pickle error when file is not top level module"
    raise unittest.SkipTest(msg)

log_file = "mp_logger_test.txt"
logger = lg.getLogger(__name__)
logger.setLevel(lg.INFO)
handler = MultiprocessHandler(log_file)
logger.addHandler(handler)


def function_logging(msg):
    """Log message via logger."""
    logger.info(msg)


args = ["Indx: {}".format(i) for i in range(32)]

workers = mp.Pool(2)
workers.map(function_logging, args)

# Try the read the logging file back and ensure that
# it has the expected format
record_found = [False for _ in range(len(args))]
with open(log_file, 'r') as infile:
    lines = infile.readlines()

for line in lines:
    # This will raise a ValueError if it does not find the line
    # in args. In that case something went wrong with the logger
    indx = args.index(line[:-1])  # Skip the trailing '\n'

os.remove(log_file)



"""Test case for tha GAFit"""
from clease import GAFit
from clease.ga_fit import SaturatedPopulationError
from clease import CEBulk, NewStructures, Concentration
from clease.tools import update_db
from ase.calculators.emt import EMT
from ase.db import connect
from random import choice
import numpy as np
import os
import unittest


def init_system(db_name):
    basis_elements = [['Au', 'Cu']]
    concentration = Concentration(basis_elements=basis_elements)
    bc_setting = CEBulk(crystalstructure='fcc', a=4.05, size=[3, 3, 3],
                        concentration=concentration, db_name=db_name,
                        max_cluster_size=3, max_cluster_dia=[0, 0, 4.06, 4.06])

    newstruct = NewStructures(bc_setting, struct_per_gen=3)
    newstruct.generate_initial_pool()

    # Insert 30 random configurations
    atoms = bc_setting.atoms.copy()
    symbols = ["Au", "Cu"]
    for _ in range(6):
        for atom in atoms:
            atom.symbol = choice(symbols)
        newstruct.insert_structure(init_struct=atoms)

    # Compute the energy of the structures
    calc = EMT()
    database = connect(db_name)

    # Write the atoms to the database
    # for atoms, kvp in zip(all_atoms, key_value_pairs):
    for row in database.select([("converged", "=", False)]):
        atoms = row.toatoms()
        atoms.set_calculator(calc)
        atoms.get_potential_energy()
        update_db(uid_initial=row.id, final_struct=atoms, db_name=db_name)
    return bc_setting


class TestGAFit(unittest.TestCase):
    def test_init_from_file(self):
        db_name = 'test_ga_fit.db'
        setting = init_system(db_name)
        backup = "ga_test.csv"
        selector = GAFit(setting=setting, fname=backup)
        try:
            selector.run(gen_without_change=3, save_interval=1)
        except SaturatedPopulationError:
            pass
        individuals = selector.individuals

        # Restart from file, this time data will be loaded
        selector = GAFit(setting=setting, fname=backup)
        self.assertTrue(np.allclose(individuals, selector.individuals))
        os.remove(backup)

        if os.path.exists("ga_test_cf_names.txt"):
            os.remove("ga_test_cf_names.txt")

    def tearDown(self):
        try:
            os.remove('test_ga_fit.db')
        except Exception:
            pass


if __name__ == '__main__':
    unittest.main()



from clease import LinearRegression
import numpy as np
import unittest


class TestRegression(unittest.TestCase):
    def test_non_singular(self):
        x = np.linspace(0.0, 1.0, 20)

        y = 1.0 + 2.0*x - 4.0*x**2

        X = np.zeros((len(x), 3))
        X[:, 0] = 1.0
        X[:, 1] = x
        X[:, 2] = x**2

        linreg = LinearRegression()
        coeff = linreg.fit(X, y)

        # Test that fit works
        self.assertTrue(np.allclose(coeff, [1.0, 2.0, -4.0]))

        # Test that precision matrix gives correct result
        # in the case where it is not singular
        prec = np.linalg.inv(X.T.dot(X))
        prec_regr = linreg.precision_matrix(X)
        self.assertTrue(np.allclose(prec, prec_regr))

    def test_trivial_singular(self):
        x = np.linspace(0.0, 1.0, 20)

        y = 1.0 + 2.0*x - 4.0*x**2

        X = np.zeros((len(x), 4))
        X[:, 0] = 1.0
        X[:, 1] = x
        X[:, 2] = x**2
        X[:, 3] = x**2

        linreg = LinearRegression()
        coeff = linreg.fit(X, y)

        self.assertTrue(np.allclose(X.dot(coeff), y))
        linreg.precision_matrix(X)

    def test_complicated_singular(self):
        x = np.linspace(0.0, 1.0, 20)

        y = 1.0 + 2.0*x - 4.0*x**2

        X = np.zeros((len(x), 5))
        X[:, 0] = 1.0
        X[:, 1] = x
        X[:, 2] = x**2
        X[:, 3] = 0.1 - 0.2*x + 0.8*x**2
        X[:, 4] = -0.2 + 0.8*x

        linreg = LinearRegression()
        coeff = linreg.fit(X, y)

        self.assertTrue(np.allclose(X.dot(coeff), y))
        linreg.precision_matrix(X)


if __name__ == '__main__':
    unittest.main()



import unittest


class ConcentrationPageTest(unittest.TestCase):

    def check_add_constraint(self, app):
        screen = app.root.ids.sm.get_screen('Concentration')
        num_children = len(screen.ids.mainConcLayout.children)
        screen.ids.addConstraintButton.dispatch('on_release')
        self.assertEqual(len(screen.ids.mainConcLayout.children),
                         num_children + 1)

        # Locate the remove button
        for widget in screen.ids.mainConcLayout.children:
            if widget.id is None:
                continue
            if widget.id == 'cnst0':
                for ch in widget.children:
                    if ch.text == 'Remove':
                        ch.dispatch('on_press')
        self.assertEqual(len(screen.ids.mainConcLayout.children),
                         num_children)

    def check_matrices(self, app):
        screen = app.root.ids.sm.get_screen('Concentration')
        screen.ids.elementInput.text = 'Al, Mg, Si'
        screen.ids.groupedBasisInput.text = ''
        screen.ids.applyElemGroupButton.dispatch('on_release')

        # Constraint counter is on 1 after the check_add_constraint
        self.assertEqual(screen.next_constraint_id, 1)
        screen.ids.addConstraintButton.dispatch('on_release')
        screen.ids.addConstraintButton.dispatch('on_release')
        screen.ids.addConstraintButton.dispatch('on_release')
        self.assertEqual(screen.next_constraint_id, 4)

        for widget in screen.ids.mainConcLayout.children:
            if widget.id is None:
                continue

            if widget.id == 'cnst1':
                for ch in widget.children:
                    if ch.id is None:
                        continue

                    if ch.id == 'conc0' or ch.id == 'conc1':
                        ch.text = '1.0'
                    elif ch.id == 'rhs':
                        ch.text = '0.5'
                    elif ch.id == 'comparisonSpinner':
                        ch.text = '='
            elif widget.id == 'cnst2':
                for ch in widget.children:
                    if ch.id is None:
                        continue

                    if ch.id == 'conc2':
                        ch.text = '1.0'
                    elif ch.id == 'rhs':
                        ch.text = '0.5'
                    elif ch.id == 'comparisonSpinner':
                        ch.text = '<='
            elif widget.id == 'cnst3':
                for ch in widget.children:
                    if ch.id is None:
                        continue

                    if ch.id == 'conc1':
                        ch.text = '1.0'
                    elif ch.id == 'rhs':
                        ch.text = '0.5'
                    elif ch.id == 'comparisonSpinner':
                        ch.text = '>='
        A_lb, rhs_lb, A_eq, rhs_eq = screen.get_constraint_matrices()

        A_lb_expect = [[0, 1, 0], [0, 0, -1]]
        rhs_lb_expect = [0.5, -0.5]
        A_eq_expect = [[1, 1, 0]]
        rhs_eq_expect = [0.5]

        for row, row_exp in zip(A_lb, A_lb_expect):
            for x, y in zip(row, row_exp):
                self.assertAlmostEqual(x, y)

        for row, row_exp in zip(A_eq, A_eq_expect):
            for x, y in zip(row, row_exp):
                self.assertAlmostEqual(x, y)

        for x, y in zip(rhs_lb, rhs_lb_expect):
            self.assertAlmostEqual(x, y)

        for x, y in zip(rhs_eq, rhs_eq_expect):
            self.assertAlmostEqual(x, y)

    def run_check_element_input(self, app):
        screen = app.root.ids.sm.get_screen('Concentration')
        elem_in = screen.ids.elementInput
        elem_in.text = 'Al, Cu'
        self.assertTrue(screen.elem_ok())
        elem_in.text = '[[Zn, Cu]]'
        self.assertTrue(screen.elem_ok())
        elem_in.text = '(Al, Cu), (Mg, Si)'
        self.assertTrue(screen.elem_ok())
        elem_in.text = '[(Al, Cu), (Mg, Si)]'
        self.assertTrue(screen.elem_ok())

    def run_check_grouped_basis(self, app):
        screen = app.root.ids.sm.get_screen('Concentration')
        gr_basis = screen.ids.groupedBasisInput
        gr_basis.text = '1, 2'
        self.assertTrue(screen.grouped_basis_ok())
        gr_basis.text = '[1, 2]'
        self.assertTrue(screen.grouped_basis_ok())
        gr_basis.text = '[1, 2], 3'
        self.assertTrue(screen.grouped_basis_ok())
        gr_basis.text = '(1, 2), [3, 4]'
        self.assertTrue(screen.grouped_basis_ok())
        gr_basis.text = '(1, 2), (3, 4)'
        self.assertTrue(screen.grouped_basis_ok())
        gr_basis.text = '[(1, 2), (3, 4)]'
        self.assertTrue(screen.grouped_basis_ok())

    def run_with_app(self, app):
        self.check_add_constraint(app)
        self.check_matrices(app)
        self.run_check_element_input(app)
        self.run_check_grouped_basis(app)



import os
import unittest
from clease import BayesianCompressiveSensing
import numpy as np
from scipy.special import polygamma

# Fix the seed to ensure consistent tests
np.random.seed(0)
fname = "test_bayes_compr_sens.json"


class TestBayesianCompressiveSensing(unittest.TestCase):
    bayes = BayesianCompressiveSensing(fname=fname, output_rate_sec=2,
                                       maxiter=100)

    def test_optimize_shape_parameter(self):
        self.bayes.lamb = 1.0
        opt = self.bayes.optimal_shape_lamb()
        self.assertAlmostEqual(np.log(opt/2.0), polygamma(0, opt/2.0))

    def test_fit(self):
        X = np.random.rand(30, 400)
        y = 60.0*X[:, 20] - 80.0*X[:, 2]
        eci = self.bayes.fit(X, y)

        expected_eci = np.zeros(X.shape[1])
        expected_eci[20] = 60.0
        expected_eci[2] = -80.0
        self.assertTrue(np.allclose(eci, expected_eci, rtol=1E-4))

    def test_fit_more_coeff(self):
        self.bayes = BayesianCompressiveSensing(fname=fname, noise=0.1,
                                                maxiter=1000)
        X = np.random.rand(30, 400)
        coeff = [6.0, -2.0, 5.0, 50.0, -30.0]
        indx = [0, 23, 19, 18, 11]
        y = 0.0
        expected_eci = np.zeros(X.shape[1])
        for c, i in zip(coeff, indx):
            y += X[:, i]*c
            expected_eci[i] = c
        eci = self.bayes.fit(X, y)
        self.assertTrue(np.allclose(eci, expected_eci, atol=1E-2))

    def test_save_load(self):
        self.bayes.save()
        bayes2 = BayesianCompressiveSensing.load(fname)
        self.assertTrue(self.bayes == bayes2)

    def test_fit_linear_dep_col(self):
        bayes = BayesianCompressiveSensing(fname=fname, noise=0.2,
                                           penalty=1E-2, maxiter=1000)
        X = np.random.rand(30, 400)
        X[:, 2] = X[:, 0]
        X[:, 8] = X[:, 20]
        X[:, 23] = X[:, 50]
        y = 20*X[:, 0] - 3*X[:, 23]
        eci = bayes.fit(X, y)
        self.assertEqual(len(eci), 400)

        expected = np.zeros(400)
        expected[0] = 20
        expected[23] = -3
        print(np.argwhere(np.abs(eci) > 1E-8))
        self.assertTrue(np.allclose(eci, expected, atol=1E-2))

        prec = bayes.precision_matrix(X)
        self.assertTrue(prec.shape == (400, 400))

    def tearDown(self):
        try:
            os.remove(fname)
        except Exception:
            pass


if __name__ == '__main__':
    unittest.main()



import numpy as np
from clease.concentration import Concentration
from collections import OrderedDict
from clease.concentration import InvalidConstraintError
import unittest


class TestConc(unittest.TestCase):
    def test_full_range(self):
        basis_elements = [['Li', 'Ru', 'X'], ['O', 'X']]
        conc_cls = Concentration(basis_elements=basis_elements)
        conc = conc_cls.get_random_concentration()
        sum1 = np.sum(conc[:3])
        self.assertAlmostEqual(sum1, 1)
        sum2 = np.sum(conc[3:])
        self.assertAlmostEqual(sum2, 1)

        # Test exceptions
        # 1) Different length for A_eq and b_eq
        A = [[1, 0, 0, 0, 1], [2, 3, 0, 1, 0]]
        b = [0]
        with self.assertRaises(InvalidConstraintError):
            conc_cls.add_usr_defined_eq_constraints(A, b)

        with self.assertRaises(InvalidConstraintError):
            conc_cls.add_usr_defined_ineq_constraints(A, b)

        # 2) Wrong number of columns
        A = [[1, 1]]
        b = [0]
        with self.assertRaises(InvalidConstraintError):
            conc_cls.add_usr_defined_eq_constraints(A, b)

        with self.assertRaises(InvalidConstraintError):
            conc_cls.add_usr_defined_ineq_constraints(A, b)

        # 3) Wrong dimension on A matrix
        A = [1, 0, 0, 0, 0]
        b = [0]
        with self.assertRaises(InvalidConstraintError):
            conc_cls.add_usr_defined_eq_constraints(A, b)

        with self.assertRaises(InvalidConstraintError):
            conc_cls.add_usr_defined_ineq_constraints(A, b)

        # 4) Wrong dimension on the b vector
        A = [[1, 0, 0, 0, 0]]
        b = [[0, 1]]
        with self.assertRaises(InvalidConstraintError):
            conc_cls.add_usr_defined_eq_constraints(A, b)

        with self.assertRaises(InvalidConstraintError):
            conc_cls.add_usr_defined_ineq_constraints(A, b)

        # 5) Wrong number of basis
        ranges = [[(0, 1), (0, 3)]]
        with self.assertRaises(InvalidConstraintError):
            conc_cls.set_conc_ranges(ranges)

        # 6) Wrong number of ranges in each basis
        ranges = [[(0, 1), (0, 0.5)], [(0, 1), (0, 1)]]
        with self.assertRaises(InvalidConstraintError):
            conc_cls.set_conc_ranges(ranges)

        # 7) Wrong bounds
        ranges = [[(0, 1), (0, 0.5), (-0.5, 2.1)], [(0, 1), (0, 1)]]
        with self.assertRaises(InvalidConstraintError):
            conc_cls.set_conc_ranges(ranges)

        # 8) Wrong number of bounds
        ranges = [[(0, 1), (0, 0.5), (0, 1, 0.5)], [(0, 1), (0, 1)]]
        with self.assertRaises(InvalidConstraintError):
            conc_cls.set_conc_ranges(ranges)

        # 9) Formula not passed
        variable_range = {"x": (0, 1)}
        with self.assertRaises(InvalidConstraintError):
            conc_cls.set_conc_formula_unit(variable_range=variable_range)

        # 10) Variable range not passed
        formulas = []
        with self.assertRaises(InvalidConstraintError):
            conc_cls.set_conc_formula_unit(formulas=formulas)

        # 11) Wrong number of formulas
        with self.assertRaises(InvalidConstraintError):
            conc_cls.set_conc_formula_unit(formulas=formulas,
                                           variable_range=variable_range)

    def fixed_composition(self):
        basis_elements = [['Li', 'Ru'], ['O', 'X']]
        A_eq = [[0, 3, 0, 0], [0, 0, 0, 2]]
        b_eq = [1, 1]
        conc = Concentration(basis_elements=basis_elements, A_eq=A_eq,
                             b_eq=b_eq)
        rand = conc.get_random_concentration()
        self.assertTrue(np.allclose(rand, np.array([2./3, 1./3, 0.5, 0.5])))

    def test_fix_Ru_composition(self):
        basis_elements = [['Li', 'Ru', 'X'], ['O', 'X']]
        A_eq = [[0, 3, 0, 0, 0]]
        b_eq = [1]
        A_lb = [[0, 0, 0, 3, 0]]
        b_lb = [2]
        conc = Concentration(basis_elements=basis_elements, A_eq=A_eq,
                             b_eq=b_eq, A_lb=A_lb, b_lb=b_lb)
        rand = conc.get_random_concentration()
        sum1 = np.sum(rand[:3])
        self.assertAlmostEqual(sum1, 1)
        sum2 = np.sum(rand[3:])
        self.assertAlmostEqual(sum2, 1)

    def test_conc_range(self):
        basis_elements = [['Li', 'Ru', 'X'], ['O', 'X']]
        conc = Concentration(basis_elements=basis_elements)
        ranges = [[(0, 1), (1./3, 1./3), (0, 1)], [(2./3, 1), (0, 1)]]
        conc.set_conc_ranges(ranges)
        rand = conc.get_random_concentration()
        sum1 = np.sum(rand[:3])
        self.assertAlmostEqual(sum1, 1)
        sum2 = np.sum(rand[3:])
        self.assertAlmostEqual(sum2, 1)
        self.assertAlmostEqual(rand[1], 1./3)
        self.assertGreater(rand[3], 2./3)

    def test_formula_unit1(self):
        basis_elements = [['Li', 'Ru', 'X'], ['O', 'X']]
        conc = Concentration(basis_elements=basis_elements)
        # test _get_integers() function
        example_str = "df<12> kd542,32"
        expected_ints = [12, 542, 32]
        self.assertTrue(conc._get_integers(example_str) == expected_ints)

    def test_formula_unit2(self):
        basis_elements = [['Li', 'Ru', 'X'], ['O', 'X']]
        formulas = ["Li<x>Ru<1>X<2-x>", "O<3-y>X<y>"]
        var_range = OrderedDict()
        var_range["x"] = (0, 2)
        var_range["y"] = (0, 0.75)
        conc = Concentration(basis_elements=basis_elements)
        conc.set_conc_formula_unit(formulas=formulas, variable_range=var_range)
        A_eq = np.array([[1, 1, 1, 0, 0],
                         [0, 0, 0, 1, 1],
                         [0, 3, 0, 0, 0]])
        b_eq = np.array([1, 1, 1])
        self.assertTrue(np.allclose(A_eq, conc.A_eq))
        self.assertTrue(np.allclose(b_eq, conc.b_eq))
        A_lb_implicit = np.identity(5)
        b_lb_implicit = np.zeros(5)
        A_lb = np.array([[3, 0, 0, 0, 0],
                         [-3, 0, 0, 0, 0],
                         [0, 0, 0, 0, 3],
                         [0, 0, 0, 0, -3]])
        b_lb = np.array([0, -2, 0, -0.75])
        A_lb = np.vstack((A_lb_implicit, A_lb))
        b_lb = np.append(b_lb_implicit, b_lb)
        self.assertTrue(np.allclose(A_lb, conc.A_lb))
        self.assertTrue(np.allclose(b_lb, conc.b_lb))

    def test_formula_unit3(self):
        basis_elements = [['Li', 'V', 'X'], ['O', 'F']]
        formulas = ["Li<x>V<1>X<2-x>", "O<2>F<1>"]
        range = OrderedDict({"x": (0, 2)})
        conc = Concentration(basis_elements=basis_elements)
        conc.set_conc_formula_unit(formulas=formulas, variable_range=range)
        A_eq = np.array([[1, 1, 1, 0, 0],
                         [0, 0, 0, 1, 1],
                         [0, 3, 0, 0, 0],
                         [0, 0, 0, 3, 0],
                         [0, 0, 0, 0, 3]])
        b_eq = np.array([1, 1, 1, 2, 1])
        self.assertTrue(np.allclose(A_eq, conc.A_eq))
        self.assertTrue(np.allclose(b_eq, conc.b_eq))
        A_lb_implicit = np.identity(5)
        b_lb_implicit = np.zeros(5)
        A_lb = np.array([[3, 0, 0, 0, 0],
                         [-3, 0, 0, 0, 0]])
        b_lb = np.array([0, -2])
        A_lb = np.vstack((A_lb_implicit, A_lb))
        b_lb = np.append(b_lb_implicit, b_lb)
        self.assertTrue(np.allclose(A_lb, conc.A_lb))
        self.assertTrue(np.allclose(b_lb, conc.b_lb))

    def test_formula_unit4(self):
        basis_elements = [['Al', 'Mg', 'Si']]
        formulas = ['Al<4-4x>Mg<3x>Si<x>']
        range = OrderedDict({"x": (0, 1)})
        conc = Concentration(basis_elements=basis_elements)
        conc.set_conc_formula_unit(formulas=formulas, variable_range=range)
        A_eq = np.array([[1, 1, 1],
                         [0, 1, -3]])
        b_eq = np.array([1, 0])
        self.assertTrue(np.allclose(A_eq, conc.A_eq))
        self.assertTrue(np.allclose(b_eq, conc.b_eq))
        A_lb_implicit = np.identity(3)
        b_lb_implicit = np.zeros(3)
        A_lb = np.array([[0, 0, 4],
                         [0, 0, -4]])
        b_lb = np.array([0, -1])
        A_lb = np.vstack((A_lb_implicit, A_lb))
        b_lb = np.append(b_lb_implicit, b_lb)
        self.assertTrue(np.allclose(A_lb, conc.A_lb))
        self.assertTrue(np.allclose(b_lb, conc.b_lb))

    def test_formula_unit5(self):
        basis_elements = [['Al', 'Mg', 'Si']]
        formulas = ['Al<2-2x>Mg<x>Si<x>']
        range = OrderedDict({"x": (0, 1)})
        conc = Concentration(basis_elements=basis_elements)
        conc.set_conc_formula_unit(formulas=formulas, variable_range=range)
        A_eq = np.array([[1, 1, 1],
                         [0, -1, 1]])
        b_eq = np.array([1, 0])
        self.assertTrue(np.allclose(A_eq, conc.A_eq))
        self.assertTrue(np.allclose(b_eq, conc.b_eq))
        A_lb_implicit = np.identity(3)
        b_lb_implicit = np.zeros(3)
        A_lb = np.array([[0, 2, 0],
                         [0, -2, 0]])
        b_lb = np.array([0, -1])
        A_lb = np.vstack((A_lb_implicit, A_lb))
        b_lb = np.append(b_lb_implicit, b_lb)
        self.assertTrue(np.allclose(A_lb, conc.A_lb))
        self.assertTrue(np.allclose(b_lb, conc.b_lb))

    def test_formula_unit6(self):
        basis_elements = [['Al', 'Mg', 'Si', 'X']]
        formulas = ['Al<3-x-2y>Mg<y>Si<y>X<x>']
        range = {"x": (0, 1), "y": (0, 1)}
        conc = Concentration(basis_elements=basis_elements)
        conc.set_conc_formula_unit(formulas=formulas, variable_range=range)
        A_eq = np.array([[1, 1, 1, 1],
                         [0, -1, 1, 0]])
        b_eq = np.array([1, 0])
        self.assertTrue(np.allclose(A_eq, conc.A_eq))
        self.assertTrue(np.allclose(b_eq, conc.b_eq))
        A_lb_implicit = np.identity(4)
        b_lb_implicit = np.zeros(4)
        A_lb = np.array([[0, 0, 0, 3],
                         [0, 0, 0, -3],
                         [0, 3, 0, 0],
                         [0, -3, 0, 0]])
        b_lb = np.array([0, -1, 0, -1])
        A_lb = np.vstack((A_lb_implicit, A_lb))
        b_lb = np.append(b_lb_implicit, b_lb)
        self.assertTrue(np.allclose(A_lb, conc.A_lb))
        self.assertTrue(np.allclose(b_lb, conc.b_lb))

    def test_three_interlinked_basis(self):
        basis_elements = [["Al", "Mg", "Si"], ["X", "O"], ["Ta", "Se"]]
        A_eq = [[1, 0, 0, -1, 0, 0, 0]]
        b_eq = [0]
        A_lb = [[0, 0, 1, 0, 0, -1, 0]]
        b_lb = [0]
        conc = Concentration(basis_elements=basis_elements,
                             A_eq=A_eq, b_eq=b_eq, A_lb=A_lb, b_lb=b_lb)
        linked = conc._linked_basis
        self.assertEqual(sum(1 for i, num in enumerate(linked) if num == i), 1)

        A_lb = [[1, 0, 0, 1, 0, -1, 0]]
        b_lb = [0]

        conc = Concentration(basis_elements=basis_elements,
                             A_lb=A_lb, b_lb=b_lb)
        linked = conc._linked_basis
        self.assertEqual(sum(1 for i, num in enumerate(linked) if num == i), 1)

    def test_two_of_four_linked_basis(self):
        basis_elements = [["Al", "Mg"], ["X", "O"], ["Ta", "Se"], ["Ta", "O"]]
        A_eq = [[1, 0, 0, 0, -1, 0, 0, 0]]
        b_eq = [0]
        A_lb = [[0, 0, 1, 0, 0, 0, 0, -1]]
        b_lb = [0]
        conc = Concentration(basis_elements=basis_elements,
                             A_eq=A_eq, b_eq=b_eq, A_lb=A_lb, b_lb=b_lb)
        linked = conc._linked_basis
        self.assertEqual(sum(1 for i, num in enumerate(linked) if num == i), 2)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import numpy as np
from clease.calculator import attach_calculator
from clease.montecarlo import Montecarlo
from clease.montecarlo.observers import CorrelationFunctionObserver
from clease.montecarlo.observers import Snapshot
from clease.montecarlo.observers import EnergyEvolution
from clease.montecarlo.observers import SiteOrderParameter
from clease.montecarlo.observers import LowestEnergyStructure
from clease.montecarlo.observers import DiffractionObserver
from clease.montecarlo.constraints import ConstrainSwapByBasis
from clease.montecarlo.constraints import FixedElement
from clease import Concentration, CEBulk, CorrFunction

# Set the random seed
np.random.seed(0)


def get_example_mc_system(db_name):
    conc = Concentration(basis_elements=[['Au', 'Cu']])
    setting = CEBulk(db_name=db_name, concentration=conc,
                     crystalstructure='fcc', a=4.0,
                     max_cluster_size=3, max_cluster_dia=[5.0, 4.1],
                     size=[2, 2, 2])

    atoms = setting.atoms.copy()*(3, 3, 3)
    cf = CorrFunction(setting)
    cf_scratch = cf.get_cf(setting.atoms)
    eci = {k: 0.0 for k, v in cf_scratch.items()}

    eci['c0'] = -1.0
    eci['c2_d0000_0_00'] = -0.2
    atoms = attach_calculator(setting, atoms=atoms, eci=eci)
    return atoms


def get_rocksalt_mc_system(db_name):
    conc = Concentration(basis_elements=[['Si', 'X'], ['O', 'C']])
    setting = CEBulk(db_name=db_name, concentration=conc,
                     crystalstructure='rocksalt', a=4.0,
                     max_cluster_size=3, max_cluster_dia=[2.51, 3.0],
                     size=[2, 2, 2])
    atoms = setting.atoms.copy()*(3, 3, 3)
    cf = CorrFunction(setting)
    cf_scratch = cf.get_cf(setting.atoms)
    eci = {k: 0.0 for k, v in cf_scratch.items()}

    eci['c0'] = -1.0
    eci['c2_d0000_0_00'] = -0.2
    atoms = attach_calculator(setting, atoms=atoms, eci=eci)
    return atoms


class TestMonteCarlo(unittest.TestCase):
    def test_run(self):
        db_name = 'mc_test_aucu.db'
        atoms = get_example_mc_system(db_name)

        # Insert a few elements
        for i in range(10):
            atoms[i].symbol = 'Cu'

        E = []
        for T in [1000, 500, 100]:
            mc = Montecarlo(atoms, T)
            mc.run(steps=10000)
            E.append(mc.get_thermodynamic_quantities()['energy'])

        cf = CorrFunction(atoms.get_calculator().setting)
        cf_calc = atoms.get_calculator().get_cf()
        cf_scratch = cf.get_cf(atoms)

        os.remove(db_name)
        for k, v in cf_scratch.items():
            self.assertAlmostEqual(v, cf_calc[k])

        # Make sure that the energies are decreasing by comparing
        # the first and last
        self.assertGreaterEqual(E[0], E[-1])

    def test_corr_func_observer(self):
        db_name = 'test_corr_func_observer.db'
        atoms = get_example_mc_system(db_name)

        atoms[0].symbol = 'Cu'
        atoms[1].symbol = 'Cu'

        mc = Montecarlo(atoms, 600)
        obs = CorrelationFunctionObserver(atoms.get_calculator())
        mc.attach(obs, interval=1)
        mc.run(steps=1000)
        thermo = mc.get_thermodynamic_quantities()
        _ = obs.get_averages()
        os.remove(db_name)
        self.assertEqual(obs.counter, 1001)

        cf_keys = atoms.get_calculator().get_cf().keys()

        for k in cf_keys:
            self.assertTrue('cf_' + k in thermo.keys())

    def test_snapshot(self):
        db_name = 'test_snapshot.db'
        atoms = get_example_mc_system(db_name)

        atoms[0].symbol = 'Cu'
        atoms[1].symbol = 'Cu'

        obs = Snapshot(fname='snapshot', atoms=atoms)

        mc = Montecarlo(atoms, 600)
        mc.attach(obs, interval=100)
        mc.run(steps=1000)
        os.remove(db_name)
        self.assertEqual(len(obs.traj), 10)
        os.remove('snapshot.traj')

    def test_energy_evolution(self):
        db_name = 'test_energy_evolution.db'

        atoms = get_example_mc_system(db_name)
        atoms[0].symbol = 'Cu'
        atoms[1].symbol = 'Cu'

        mc = Montecarlo(atoms, 600)
        obs = EnergyEvolution(mc)
        mc.attach(obs, interval=50)
        mc.run(steps=1000)

        # Just confirm that the save function works
        obs.save(fname='energy_evol')
        os.remove('energy_evol.csv')

        # Check the number of energy values
        os.remove(db_name)
        self.assertEqual(20, len(obs.energies))

    def test_site_order_parameter(self):
        db_name = 'test_site_order.db'
        atoms = get_example_mc_system(db_name)

        atoms[0].symbol = 'Cu'
        atoms[1].symbol = 'Cu'
        atoms[2].symbol = 'Cu'

        obs = SiteOrderParameter(atoms)
        mc = Montecarlo(atoms, 600)
        mc.attach(obs)
        mc.run(steps=1000)
        avg = obs.get_averages()
        self.assertLessEqual(avg['site_order_average'], 6.0)

        thermo = mc.get_thermodynamic_quantities()

        os.remove(db_name)
        self.assertTrue('site_order_average' in thermo.keys())
        self.assertTrue('site_order_std' in thermo.keys())

    def test_lowest_energy_obs(self):
        db_name = 'test_low_energy.db'
        atoms = get_example_mc_system(db_name)

        atoms[0].symbol = 'Cu'
        atoms[1].symbol = 'Cu'
        atoms[2].symbol = 'Cu'

        low_en = LowestEnergyStructure(atoms)

        mc = Montecarlo(atoms, 700)
        energy_evol = EnergyEvolution(mc)
        mc.attach(low_en, interval=1)
        mc.attach(energy_evol, interval=1)

        mc.run(steps=1000)
        os.remove(db_name)
        self.assertAlmostEqual(np.min(energy_evol.energies),
                               low_en.lowest_energy + mc.energy_bias)

    def test_diffraction_obs(self):
        db_name = 'test_diffraction_observer.db'
        atoms = get_example_mc_system(db_name)

        atoms[0].symbol = 'Cu'
        atoms[1].symbol = 'Cu'
        atoms[2].symbol = 'Cu'

        obs = DiffractionObserver(atoms=atoms, k_vector=[0.25, 0.0, 0.0],
                                  name='reflect1', active_symbols=['Cu'],
                                  all_symbols=['Au', 'Cu'])

        mc = Montecarlo(atoms, 600)
        mc.attach(obs)

        mc.run(steps=1000)
        thermo = mc.get_thermodynamic_quantities()

        os.remove(db_name)
        self.assertTrue('reflect1' in thermo.keys())

    def test_constrain_swap(self):
        db_name = 'test_constrain_swap.db'
        atoms = get_rocksalt_mc_system(db_name)
        setting = atoms.get_calculator().setting
        i_by_basis = setting.index_by_basis

        # Insert a few vacancies
        num_X = 0
        for atom in atoms:
            if atom.symbol == 'Si':
                atom.symbol = 'X'
                num_X += 1

            if num_X >= 20:
                break

        # Insert a few C
        num_C = 0
        for atom in atoms:
            if atom.symbol == 'O':
                atom.symbol = 'C'
                num_C += 1

            if num_C >= 20:
                break

        orig_symbs = [atom.symbol for atom in atoms]

        cnst = ConstrainSwapByBasis(atoms, i_by_basis)

        mc = Montecarlo(atoms, 600)
        mc.add_constraint(cnst)
        mc.run(steps=1000)

        # Confirm that swaps have been made
        symbs = [atom.symbol for atom in atoms]
        os.remove(db_name)
        self.assertFalse(all(map(lambda x: x[0] == x[1],
                                 zip(orig_symbs, symbs))))

        allowed_elements = [['Si', 'X'], ['O', 'C']]
        for basis, allowed in zip(i_by_basis, allowed_elements):
            for indx in basis:
                self.assertTrue(atoms[indx].symbol in allowed)

    def test_fixed_element(self):
        db_name = 'test_fixed_element_constraint.db'
        atoms = get_rocksalt_mc_system(db_name)

        # Insert a few vacancies
        num_X = 0
        for atom in atoms:
            if atom.symbol == 'Si':
                atom.symbol = 'X'
                num_X += 1

            if num_X >= 20:
                break

        # Insert a few C
        num_C = 0
        for atom in atoms:
            if atom.symbol == 'O':
                atom.symbol = 'C'
                num_C += 1

            if num_C >= 20:
                break

        # Let's say that all Si atoms should be fixed
        fixed_element = FixedElement('Si')

        mc = Montecarlo(atoms, 10000)
        mc.add_constraint(fixed_element)

        si_indices = [atom.index for atom in atoms if atom.symbol == 'Si']
        mc.run(steps=1000)

        si_after = [atom.index for atom in atoms if atom.symbol == 'Si']
        os.remove(db_name)
        self.assertEqual(si_indices, si_after)


if __name__ == '__main__':
    unittest.main()



import unittest
from clease.atoms_manager import AtomsManager
from ase.build import bulk
from ase.spacegroup import crystal
import numpy as np


class TestAtomsManager(unittest.TestCase):
    def test_binary(self):
        atoms = bulk('Au')*(3, 3, 3)

        # Tag even indices with 0 and odd indices with 1
        for atom in atoms:
            atom.tag = atom.index % 2

        manager = AtomsManager(atoms)
        ind_by_tag = manager.index_by_tag()
        self.assertTrue(all(map(lambda x: x % 2 == 0, ind_by_tag[0])))
        self.assertTrue(all(map(lambda x: x % 2 == 1, ind_by_tag[1])))

    def test_group_by_symbol_single(self):
        atoms = bulk('Au')*(3, 3, 3)

        for atom in atoms:
            if atom.index % 3 == 1:
                atom.symbol = 'Cu'
            elif atom.index % 3 == 2:
                atom.symbol = 'X'

        manager = AtomsManager(atoms)
        ind_by_symbol = manager.index_by_symbol(['Au', 'Cu', 'X'])

        for i, items in enumerate(ind_by_symbol):
            self.assertTrue(all(map(lambda x: x % 3 == i, items)))

    def test_group_by_symbol_grouped(self):
        atoms = bulk('Au')*(3, 4, 5)

        for atom in atoms:
            if atom.index % 4 == 1:
                atom.symbol = 'Cu'
            elif atom.index % 4 == 2:
                atom.symbol = 'X'
            elif atom.index % 4 == 3:
                atom.symbol = 'Ag'

        manager = AtomsManager(atoms)
        ind_by_symbol = manager.index_by_symbol(['Au', ['Cu', 'X'], 'Ag'])

        self.assertTrue(all(map(lambda x: x % 4 == 0, ind_by_symbol[0])))
        self.assertTrue(all(map(lambda x: x % 4 == 1 or x % 4 == 2,
                                ind_by_symbol[1])))
        self.assertTrue(all(map(lambda x: x % 4 == 3, ind_by_symbol[2])))
        self.assertEqual(sorted(manager.unique_elements()),
                         ['Ag', 'Au', 'Cu', 'X'])

    def test_background_indices(self):
        atoms = bulk('NaCl', crystalstructure='rocksalt', a=4.0)
        atoms = atoms*(5, 5, 5)

        # Chlorine sites are background indices
        basis_elements = [['Na', 'X'], ['Cl']]
        manager = AtomsManager(atoms)
        bkg_indices = manager.single_element_sites(basis_elements)

        cl_sites = [atom.index for atom in atoms if atom.symbol == 'Cl']
        self.assertEqual(sorted(bkg_indices), sorted(cl_sites))

        # Extract unique elements
        unique_elem = manager.unique_elements()
        self.assertEqual(sorted(unique_elem), ['Cl', 'Na'])

        # Try unique elements without background
        unique_elem = manager.unique_elements(ignore=['Cl'])
        self.assertEqual(sorted(unique_elem), ['Na'])

    def test_tag_by_corresponding_atom(self):
        prim_cell = bulk('Mg', crystalstructure='hcp')
        prim_cell[0].symbol = 'Mg'
        prim_cell[1].symbol = 'Zn'

        atoms = prim_cell*(2, 3, 4)
        manager = AtomsManager(atoms)
        manager.tag_indices_of_corresponding_atom(prim_cell)

        for atom in manager.atoms:
            if atom.symbol == 'Mg':
                self.assertEqual(atom.tag, 0)
            else:
                self.assertEqual(atom.tag, 1)

    def test_tag_by_corresponding_primitive_conventional(self):
        prim_cell = bulk('NaCl', crystalstructure='rocksalt', a=4.0)
        prim_cell.wrap()
        atoms = bulk('NaCl', crystalstructure='rocksalt', a=4.0, cubic=True)
        atoms = atoms*(3, 4, 5)

        manager = AtomsManager(atoms)
        manager.tag_indices_of_corresponding_atom(prim_cell)

        for atom in manager.atoms:
            if atom.symbol == 'Na':
                self.assertEqual(atom.tag, 0)
            elif atom.symbol == 'Cl':
                self.assertEqual(atom.tag, 1)

    def test_raise_if_not_match(self):
        prim_cell = bulk('NaCl', crystalstructure='rocksalt', a=4.0)
        prim_cell.wrap()
        atoms = bulk('Mg', crystalstructure='hcp')
        atoms = atoms*(3, 4, 5)

        manager = AtomsManager(atoms)
        with self.assertRaises(ValueError):
            manager.tag_indices_of_corresponding_atom(prim_cell)

    def test_cubic_sc_fcc(self):
        a = 4.05
        atoms = bulk("Al", a=a, crystalstructure="fcc")
        manager = AtomsManager(atoms)
        sc = manager.close_to_cubic_supercell()
        expected_cell = np.array([[a, 0, 0], [0, a, 0], [0, 0, a]])
        self.assertTrue(np.allclose(expected_cell, sc.get_cell()))

    def test_sp217(self):
        a = 10.553
        b = 10.553
        c = 10.553
        alpha = 90
        beta = 90
        gamma = 90
        cellpar = [a, b, c, alpha, beta, gamma]
        basis = [(0, 0, 0), (0.324, 0.324, 0.324),
                 (0.3582, 0.3582, 0.0393), (0.0954, 0.0954, 0.2725)]

        atoms = crystal(symbols=['Mg', 'Mg', 'Mg', 'Al'],
                        cellpar=cellpar, spacegroup=217,
                        basis=basis, primitive_cell=True)

        manager = AtomsManager(atoms)
        sc = manager.close_to_cubic_supercell()
        expected_cell = np.array([[a, 0, 0], [0, a, 0], [0, 0, a]])
        self.assertTrue(np.allclose(expected_cell, sc.get_cell()))

    def test_bcc(self):
        a = 3.8
        atoms = bulk("Fe", a=a, crystalstructure="bcc")
        manager = AtomsManager(atoms)
        sc = manager.close_to_cubic_supercell()
        expected_cell = np.array([[a, 0, 0], [0, a, 0], [0, 0, a]])
        self.assertTrue(np.allclose(sc.get_cell(), expected_cell))


if __name__ == '__main__':
    unittest.main()



import os
from clease import CEBulk, NewStructures, Concentration
from ase.build import bulk
from clease.tools import wrap_and_sort_by_position
import unittest


class TestFormulaUnit(unittest.TestCase):
    def test_fu(self):
        db_name = 'fu.db'
        basis_elements = [['Li', 'Ru', 'X'], ['O', 'X']]
        A_eq = [[0, 3, 0, 0, 0]]
        b_eq = [1]
        A_lb = [[0, 0, 0, 3, 0]]
        b_lb = [2]
        conc = Concentration(basis_elements=basis_elements, A_eq=A_eq,
                             b_eq=b_eq, A_lb=A_lb, b_lb=b_lb)
        setting = CEBulk(crystalstructure="rocksalt",
                         a=4.0,
                         size=[2, 2, 3],
                         concentration=conc,
                         db_name=db_name,
                         max_cluster_size=3,
                         max_cluster_dia=4.)
        newstruct = NewStructures(setting=setting)

        test = bulk(name='LiO', crystalstructure='rocksalt', a=4.0) * (2, 2, 3)
        atoms = wrap_and_sort_by_position(test.copy())
        atoms[0].symbol = 'Ru'
        atoms[1].symbol = 'X'
        atoms[4].symbol = 'X'
        fu = newstruct._get_formula_unit(atoms)
        self.assertEqual(fu, "Li10Ru1X1_O11X1")

        atoms = wrap_and_sort_by_position(test.copy())
        fu = newstruct._get_formula_unit(atoms)
        self.assertEqual(fu, "Li1_O1")

        atoms = wrap_and_sort_by_position(test.copy())
        replace_cat = [0, 1, 2, 5, 6, 14]
        replace_an = [4, 7, 9, 13, 15, 20]
        for cat in replace_cat:
            atoms[cat].symbol = 'X'
        for an in replace_an:
            atoms[an].symbol = 'X'

        fu = newstruct._get_formula_unit(atoms)
        self.assertEqual(fu, "Li1X1_O1X1")
        os.remove(db_name)


if __name__ == '__main__':
    unittest.main()



import os
from clease import CEBulk, CECrystal, Concentration
import unittest


class TestSupercelLFactor(unittest.TestCase):
    def test_fcc(self):
        db_name = 'test_sc_factor_fcc_sf.db'
        conc = Concentration(basis_elements=[['Au', 'Cu']])

        setting = CEBulk(crystalstructure='fcc', a=4.0,
                         supercell_factor=8,
                         concentration=conc, db_name=db_name,
                         max_cluster_size=4, max_cluster_dia=4.01,
                         basis_function='polynomial', skew_threshold=4,
                         ignore_background_atoms=False)

        self.assertEqual(setting.template_atoms.num_templates, 3)

        os.remove(db_name)

        setting = CEBulk(crystalstructure='fcc', a=4.01,
                         supercell_factor=8, size=[2, 2, 2],
                         concentration=conc, db_name=db_name,
                         max_cluster_size=4, max_cluster_dia=4.0,
                         basis_function='polynomial', skew_threshold=4,
                         ignore_background_atoms=False)

        self.assertEqual(setting.template_atoms.num_templates, 1)

        os.remove(db_name)

    def test_crystal(self):
        db_name = 'test_sc_factor_crystal_sf.db'
        basis_elements = [['O', 'X'], ['O', 'X'],
                          ['O', 'X'], ['Ta']]
        grouped_basis = [[0, 1, 2], [3]]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=grouped_basis)

        setting = CECrystal(basis=[(0., 0., 0.),
                                   (0.3894, 0.1405, 0.),
                                   (0.201, 0.3461, 0.5),
                                   (0.2244, 0.3821, 0.)],
                            spacegroup=55,
                            cellpar=[6.25, 7.4, 3.83, 90, 90, 90],
                            supercell_factor=10,
                            concentration=concentration,
                            db_name=db_name,
                            basis_function='binary-linear',
                            max_cluster_size=3,
                            max_cluster_dia=3.0)

        self.assertTrue(setting.template_atoms.num_templates, 20)
        os.remove(db_name)


if __name__ == '__main__':
    unittest.main()



"""Test to initiatialize CE using a CEBulk.

1. Initialize the CE
2. Add a few structures
3. Compute the energy
4. Run the evaluation routine
"""

import os
import json
from clease import CEBulk, CorrFunction, NewStructures, Evaluate, Concentration
from clease.newStruct import MaxAttemptReachedError
from clease.tools import update_db
from ase.calculators.emt import EMT
from ase.db import connect
from ase.build import bulk
from reference_corr_funcs_bulk import all_cf
from ase.build import make_supercell
import numpy as np
import unittest
from unittest.mock import patch

# If this is True, the JSON file containing the correlation functions
# Used to check consistency of the reference functions is updated
# This should normally be False
update_reference_file = False
tol = 1E-9

def get_figures_of_family(setting, cname):
    """Return the figures of a given cluster family."""
    figures = []
    clusters = setting.cluster_list.get_by_name(cname)
    for cluster in clusters:
        figures.append(cluster.indices)
    return figures

def calculate_cf(setting, atoms):
    cf = CorrFunction(setting)
    cf_dict = cf.get_cf(atoms)
    return cf_dict


class TestCEBulk(unittest.TestCase):
    def test_load_from_db(self):
        db_name = 'test_load_from_db.db'
        basis_elements = [['Au', 'Cu']]
        concentration = Concentration(basis_elements=basis_elements)
        setting = CEBulk(crystalstructure='fcc', a=4.05, size=[1, 1, 1],
                         concentration=concentration, db_name=db_name,
                         max_cluster_dia=[4.3, 4.3, 4.3],
                         max_cluster_size=4)
        orig_atoms = setting.atoms.copy()
        atoms = bulk('Au', crystalstructure='fcc', a=4.05, cubic=True)
        setting.set_active_template(atoms=atoms, generate_template=True)

        # Try to read back the old atoms
        setting.set_active_template(atoms=orig_atoms)
        os.remove(db_name)

    def test_corrfunc(self):
        db_name = "test_bulk_corrfunc.db"
        basis_elements = [['Au', 'Cu']]
        concentration = Concentration(basis_elements=basis_elements)
        setting = CEBulk(crystalstructure='fcc', a=4.05, size=[3, 3, 3],
                         concentration=concentration, db_name=db_name,
                         max_cluster_dia=[4.3, 4.3, 4.3],
                         max_cluster_size=4)
        atoms = setting.atoms.copy()
        atoms[0].symbol = 'Cu'
        atoms[3].symbol = 'Cu'
        cf = calculate_cf(setting, atoms)

        if update_reference_file:
            all_cf["binary_fcc"] = cf
        for key in cf.keys():
            self.assertAlmostEqual(cf[key], all_cf["binary_fcc"][key])

        os.remove(db_name)

        basis_elements = [['Li', 'V'], ['X', 'O']]
        concentration = Concentration(basis_elements=basis_elements)
        setting = CEBulk(crystalstructure="rocksalt",
                         a=4.0,
                         size=[2, 2, 1],
                         concentration=concentration,
                         db_name=db_name,
                         max_cluster_size=3,
                         max_cluster_dia=[4.01, 4.01])
        atoms = setting.atoms.copy()
        Li_ind = [atom.index for atom in atoms if atom.symbol == 'Li']
        X_ind = [atom.index for atom in atoms if atom.symbol == 'X']
        atoms[Li_ind[0]].symbol = 'V'
        atoms[X_ind[0]].symbol = 'O'
        cf = calculate_cf(setting, atoms)
        if update_reference_file:
            all_cf["two_basis"] = cf
        for key in cf.keys():
            self.assertAlmostEqual(cf[key], all_cf["two_basis"][key])
        os.remove(db_name)

        basis_elements = [['Na', 'Cl'], ['Na', 'Cl']]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=[[0, 1]])
        setting = CEBulk(crystalstructure="rocksalt",
                         a=4.0,
                         size=[2, 2, 1],
                         concentration=concentration,
                         db_name=db_name,
                         max_cluster_size=3,
                         max_cluster_dia=[4.01, 4.01])
        atoms = setting.atoms.copy()
        atoms[1].symbol = 'Cl'
        atoms[7].symbol = 'Cl'
        cf = calculate_cf(setting, atoms)
        if update_reference_file:
            all_cf["one_grouped_basis"] = cf
        for key in cf.keys():
            self.assertAlmostEqual(cf[key], all_cf["one_grouped_basis"][key])
        os.remove(db_name)

        basis_elements = [['Ca'], ['O', 'F'], ['O', 'F']]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=[[0], [1, 2]])
        setting = CEBulk(crystalstructure="fluorite",
                         a=4.0,
                         size=[2, 2, 2],
                         concentration=concentration,
                         db_name=db_name,
                         max_cluster_size=3,
                         max_cluster_dia=[4.01, 4.01],
                         ignore_background_atoms=True)
        atoms = setting.atoms.copy()
        O_ind = [atom.index for atom in atoms if atom.symbol == 'O']
        atoms[O_ind[0]].symbol = 'F'
        atoms[O_ind[1]].symbol = 'F'
        cf = calculate_cf(setting, atoms)
        if update_reference_file:
            all_cf["two_grouped_basis_bckgrnd"] = cf

        for key in cf.keys():
            self.assertAlmostEqual(
                cf[key], all_cf["two_grouped_basis_bckgrnd"][key]
            )
        os.remove(db_name)

    def test_binary_system(self):
        """Verifies that one can run a CE for the binary Au-Cu system.

        The EMT calculator is used for energy calculations
        """
        db_name = "test_bulk_binary_system.db"
        basis_elements = [['Au', 'Cu']]
        concentration = Concentration(basis_elements=basis_elements)
        bc_setting = CEBulk(crystalstructure='fcc', a=4.05, size=[3, 3, 3],
                            concentration=concentration,
                            db_name=db_name)

        newstruct = NewStructures(bc_setting, struct_per_gen=3)
        newstruct.generate_initial_pool()

        # Compute the energy of the structures
        calc = EMT()
        database = connect(db_name)

        # Write the atoms to the database
        # for atoms, kvp in zip(all_atoms, key_value_pairs):
        for row in database.select([("converged", "=", False)]):
            atoms = row.toatoms()
            atoms.set_calculator(calc)
            atoms.get_potential_energy()
            update_db(uid_initial=row.id, final_struct=atoms, db_name=db_name)
        # Evaluate
        Evaluate(bc_setting, fitting_scheme="l2", alpha=1E-6)

        # Test subclusters for pairs
        for cluster in bc_setting.cluster_list.get_by_size(2):
            sub_cl = bc_setting.cluster_list.get_subclusters(cluster)
            sub_cl_name = set([c.name for c in sub_cl])
            self.assertTrue(sub_cl_name == set(["c0", "c1"]))


        # Test a few known clusters. Triplet nearest neighbour
        name = "c3_d0000_0"
        triplet = bc_setting.cluster_list.get_by_name(name)[0]
        sub_cl = bc_setting.cluster_list.get_subclusters(triplet)
        sub_cl_name = set([c.name for c in sub_cl])
        self.assertTrue(sub_cl_name == set(["c0", "c1", "c2_d0000_0"]))

        name = "c3_d0001_0"
        triplet = bc_setting.cluster_list.get_by_name(name)[0]
        sub_cl = (bc_setting.cluster_list.get_subclusters(triplet))
        sub_cl_name = set([c.name for c in sub_cl])
        self.assertTrue(sub_cl_name == set(["c0", "c1", "c2_d0000_0", "c2_d0001_0"]))

        name = "c4_d0000_0"
        quad = bc_setting.cluster_list.get_by_name(name)[0]
        sub_cl = bc_setting.cluster_list.get_subclusters(quad)
        sub_cl_name = set([c.name for c in sub_cl])
        self.assertTrue(sub_cl_name == set(["c0", "c1", "c2_d0000_0", "c3_d0000_0"]))

        # Try to insert an atoms object with a strange
        P = [[-1, 1, 1], [1, -1, 1], [1, 1, -1]]
        self.assertGreater(np.linalg.det(P), 0)
        atoms = make_supercell(bc_setting.prim_cell, P)

        atoms[0].symbol = 'Cu'
        newstruct.insert_structure(init_struct=atoms, generate_template=True)

        # Test that the generate with many templates works
        newstruct.generate_gs_structure_multiple_templates(
            num_prim_cells=16, num_steps_per_temp=100,
            eci={'c0': 1.0, 'c3_d0000_0_000': -0.1}, num_templates=2)
        os.remove(db_name)

    def test_initial_pool(self):
        db_name = "test_bulk_initial_pool.db"
        basis_elements = [['Li', 'V'], ['X', 'O']]
        concentration = Concentration(basis_elements=basis_elements)

        setting = CEBulk(crystalstructure="rocksalt",
                         a=4.0,
                         size=[2, 2, 1],
                         concentration=concentration,
                         db_name=db_name,
                         max_cluster_size=3,
                         max_cluster_dia=[4.0, 4.0])
        ns = NewStructures(setting=setting, struct_per_gen=2)
        ns.generate_initial_pool()

        # At this point there should be the following
        # structures in the DB
        expected_names = ["V1_O1_0", "Li1_X1_0",
                          "V1_X1_0", "Li1_O1_0"]
        db = connect(db_name)
        for name in expected_names:
            num = sum(1 for row in db.select(name=name))
            self.assertEqual(num, 1)
        os.remove(db_name)

    def test_1grouped_basis_probe(self):
        """Test a case where a grouped_basis is used with supercell."""
        # ------------------------------- #
        # 1 grouped basis                 #
        # ------------------------------- #
        # initial_pool + probe_structures #
        # ------------------------------- #
        db_name = "test_bulk_1grouped_probe.db"
        basis_elements = [['Na', 'Cl'], ['Na', 'Cl']]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=[[0, 1]])
        setting = CEBulk(crystalstructure="rocksalt",
                         a=4.0,
                         size=[2, 2, 1],
                         concentration=concentration,
                         db_name=db_name,
                         max_cluster_size=3,
                         max_cluster_dia=[4.0, 4.0])

        self.assertEqual(setting.num_basis, 1)
        self.assertEqual(len(setting.index_by_basis), 1)
        self.assertTrue(setting.spin_dict == {'Cl': 1.0, 'Na': -1.0})
        self.assertEqual(len(setting.basis_functions), 1)
        try:
            ns = NewStructures(setting=setting, struct_per_gen=2)
            ns.generate_random_structures()
            ns.generate_initial_pool()
            ns = NewStructures(setting=setting, struct_per_gen=2)
            ns.generate_probe_structure(init_temp=1.0, final_temp=0.001,
                                        num_temp=5, num_steps_per_temp=100,
                                        approx_mean_var=True)

        except MaxAttemptReachedError as exc:
            print(str(exc))

        os.remove(db_name)

    def test_2grouped_basis_probe(self):
        # ------------------------------- #
        # 2 grouped basis                 #
        # ------------------------------- #
        # initial_pool + probe_structures #
        # ------------------------------- #
        db_name = "test_bulk_2grouped_probe.db"
        basis_elements = [['Zr', 'Ce'], ['O'], ['O']]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=[[0], [1, 2]])
        setting = CEBulk(crystalstructure="fluorite",
                         a=4.0,
                         size=[2, 2, 3],
                         concentration=concentration,
                         db_name=db_name,
                         max_cluster_size=2,
                         max_cluster_dia=[4.01])
        fam_figures = get_figures_of_family(setting, "c2_d0005_0")
        self.assertEqual(len(fam_figures[0]), 6)
        self.assertEqual(len(fam_figures[1]), 6)
        self.assertEqual(len(fam_figures[2]), 6)
        self.assertEqual(setting.num_basis, 2)
        self.assertEqual(len(setting.index_by_basis), 2)
        self.assertTrue(setting.spin_dict == {'Ce': 1.0, 'O': -1.0, 'Zr': 0})
        self.assertEqual(len(setting.basis_functions), 2)

        try:
            ns = NewStructures(setting=setting, struct_per_gen=2)
            ns.generate_initial_pool()
            ns = NewStructures(setting=setting, struct_per_gen=2)
            ns.generate_probe_structure(init_temp=1.0, final_temp=0.001,
                                        num_temp=5, num_steps_per_temp=100,
                                        approx_mean_var=True)

        except MaxAttemptReachedError as exc:
            print(str(exc))

        # Try to create a cell with previously failing size
        size = np.array([[-1, 1, 1], [1, -1, 1], [1, 1, -1]])
        atoms = make_supercell(setting.prim_cell, size)

        # This will fail if coordinatation number is wrong
        setting.set_active_template(atoms=atoms, generate_template=True)
        os.remove(db_name)

    def test_2grouped_basis_bckgrnd_probe(self):
        # ---------------------------------- #
        # 2 grouped_basis + background atoms #
        # ---------------------------------- #
        # initial_pool + probe_structures    #
        # ---------------------------------- #
        db_name = "test_bulk_2grouped_bck_probe.db"
        basis_elements = [['Ca'], ['O', 'F'], ['O', 'F']]
        concentration = Concentration(basis_elements=basis_elements,
                                      grouped_basis=[[0], [1, 2]])
        setting = CEBulk(crystalstructure="fluorite",
                         a=4.0,
                         size=[2, 2, 2],
                         concentration=concentration,
                         db_name=db_name,
                         max_cluster_size=3,
                         max_cluster_dia=[4.01, 4.01],
                         ignore_background_atoms=True)
        self.assertEqual(setting.num_basis, 2)
        self.assertEqual(len(setting.index_by_basis), 2)
        self.assertTrue(setting.spin_dict == {'F': 1.0, 'O': -1.0})
        self.assertEqual(len(setting.basis_functions), 1)

        try:
            ns = NewStructures(setting=setting, struct_per_gen=2)
            ns.generate_initial_pool()
            ns = NewStructures(setting=setting, struct_per_gen=2)
            ns.generate_probe_structure(init_temp=1.0, final_temp=0.001,
                                        num_temp=5, num_steps_per_temp=100,
                                        approx_mean_var=True)

        except MaxAttemptReachedError as exc:
            print(str(exc))

        os.remove(db_name)

    @patch('test_cebulk.CEBulk._read_data')
    @patch('test_cebulk.CEBulk._store_data')
    @patch('test_cebulk.CEBulk.create_cluster_list_and_trans_matrix')
    def test_fcc_binary_fixed_conc(self, *args):
        # c_Au = 1/3 and c_Cu = 2/3
        A_eq = [[2, -1]]
        b_eq = [0]
        db_name = 'test_fcc_binary_fixed_conc.db'
        conc = Concentration(basis_elements=[['Au', 'Cu']],
                             A_eq=A_eq, b_eq=b_eq)
        setting = CEBulk(crystalstructure='fcc', a=3.8, supercell_factor=27,
                         max_cluster_dia=5.0, max_cluster_size=3,
                         concentration=conc,
                         db_name=db_name)

        # Loop through templates and check that all satisfy constraints
        for atoms in setting.template_atoms.templates['atoms']:
            num = len(atoms)
            ratio = num/3.0
            self.assertAlmostEqual(ratio, int(ratio))
        os.remove(db_name)

    @patch('test_cebulk.CEBulk._read_data')
    @patch('test_cebulk.CEBulk._store_data')
    @patch('test_cebulk.CEBulk.create_cluster_list_and_trans_matrix')
    def test_rocksalt_conc_fixed_one_basis(self, *args):
        db_name = 'test_rocksalt_fixed_one_basis.db'
        basis_elem = [['Li', 'X'], ['O', 'F']]
        A_eq = [[0, 0, 3, -2]]
        b_eq = [0]
        conc = Concentration(basis_elements=basis_elem,
                             A_eq=A_eq, b_eq=b_eq)
        setting = CEBulk(crystalstructure='rocksalt', a=3.8,
                         supercell_factor=27, max_cluster_dia=5.0,
                         max_cluster_size=3, concentration=conc,
                         db_name=db_name)

        # Loop through and check that num_O sites is divisible by 5
        for atoms in setting.template_atoms.templates['atoms']:
            num_O = sum(1 for atom in atoms if atom.symbol == 'O')
            ratio = num_O/5.0
            self.assertAlmostEqual(ratio, int(ratio))
        os.remove(db_name)

    def tearDown(self):
        if update_reference_file:
            print("Updating the reference correlation function file")
            print("This should normally not be done.")
            with open("reference_corr_funcs_bulk.py", 'w') as outfile:
                json.dump(all_cf, outfile, indent=2, separators=(',', ': '))
        return super().tearDown()


if __name__ == '__main__':
    unittest.main()



"""Test to ensure the orthnormality of the basis functions."""
import os
import itertools
from clease import CEBulk, Concentration
import unittest


tol = 1E-9


def test_2(basis_function, db_name):
    """Test for 2 element case."""
    basis_elements = [['Au', 'Cu']]
    concentration = Concentration(basis_elements=basis_elements)

    setting = CEBulk(crystalstructure="fcc",
                     a=4.05,
                     size=[3, 3, 3],
                     concentration=concentration,
                     max_cluster_size=2,
                     db_name=db_name,
                     basis_function=basis_function)
    check_orthonormal(setting)


def test_3(basis_function, db_name):
    """Test for 3 element case."""
    basis_elements = [['Au', 'Cu', 'Ag']]
    concentration = Concentration(basis_elements=basis_elements)
    setting = CEBulk(crystalstructure="fcc",
                     a=4.05,
                     size=[3, 3, 3],
                     concentration=concentration,
                     max_cluster_size=2,
                     db_name=db_name,
                     basis_function=basis_function)
    check_orthonormal(setting)


def test_4(basis_function, db_name):
    """Test for 4 element case."""
    basis_elements = [['Au', 'Cu', 'Ag', 'Ni']]
    concentration = Concentration(basis_elements=basis_elements)
    setting = CEBulk(crystalstructure="fcc",
                     a=4.05,
                     size=[3, 3, 3],
                     concentration=concentration,
                     max_cluster_size=2,
                     db_name=db_name,
                     basis_function=basis_function)
    check_orthonormal(setting)


def test_5(basis_function, db_name):
    """Test for 5 element case."""
    basis_elements = [['Au', 'Cu', 'Ag', 'Ni', 'Fe']]
    concentration = Concentration(basis_elements=basis_elements)
    setting = CEBulk(crystalstructure="fcc",
                     a=4.05,
                     size=[3, 3, 3],
                     concentration=concentration,
                     max_cluster_size=2,
                     db_name=db_name,
                     basis_function=basis_function)
    check_orthonormal(setting)


def test_6(basis_function, db_name):
    """Test for 6 element case."""
    basis_elements = [['Au', 'Cu', 'Ag', 'Ni', 'Fe', 'H']]
    concentration = Concentration(basis_elements=basis_elements)
    setting = CEBulk(crystalstructure="fcc",
                     a=4.05,
                     size=[3, 3, 3],
                     concentration=concentration,
                     max_cluster_size=2,
                     db_name=db_name,
                     basis_function=basis_function)
    check_orthonormal(setting)


def check_orthonormal(setting):
    """Check orthonormality."""
    for bf in setting.basis_functions:
        sum = 0
        for key, _ in setting.spin_dict.items():
            sum += bf[key] * bf[key]
        sum /= setting.num_unique_elements
        assert abs(sum - 1.0) < tol

    # Check zeros
    alpha = list(range(len(setting.basis_functions)))
    comb = list(itertools.combinations(alpha, 2))
    for c in comb:
        sum = 0
        for key, _ in setting.spin_dict.items():
            sum += setting.basis_functions[c[0]][key] \
                * setting.basis_functions[c[1]][key]
        sum /= setting.num_unique_elements
        assert abs(sum) < tol


basis_function = 'polynomial'

bfs = ['polynomial', 'trigonometric']


class TestOrthonormal(unittest.TestCase):
    def test_2(self):
        db_name = 'test2.db'
        for bf in bfs:
            test_2(bf, db_name)
        os.remove(db_name)

    def test_3(self):
        db_name = 'test3.db'
        for bf in bfs:
            test_3(bf, db_name)
        os.remove(db_name)

    def test_4(self):
        db_name = 'test4.db'
        for bf in bfs:
            test_4(bf, db_name)
        os.remove(db_name)

    def test_5(self):
        db_name = 'test5.db'
        for bf in bfs:
            test_5(bf, db_name)
        os.remove(db_name)

    def test_6(self):
        db_name = 'test6.db'
        for bf in bfs:
            test_6(bf, db_name)
        os.remove(db_name)


if __name__ == '__main__':
    unittest.main()



from __future__ import division
from clease import ConvexHull
from ase.calculators.singlepoint import SinglePointCalculator
from ase.db import connect
from ase.build import bulk
import numpy as np
import os
import unittest


class TestConvexHull(unittest.TestCase):
    def test_binary(self):
        # from matplotlib import pyplot as plt
        # plt.switch_backend("agg")

        db_name = "test_binary_cnv_hull.db"
        db = connect(db_name)

        # Create energies that we know are on the
        # convex hull
        cnv_hull_enegies = [-x*(8-x) - x + 0.2 for x in range(9)]

        for n_cu in range(9):
            atoms = bulk("Au")
            atoms = atoms*(2, 2, 2)

            for i in range(n_cu):
                atoms[i].symbol = "Cu"

            calc = SinglePointCalculator(atoms, energy=cnv_hull_enegies[n_cu])
            atoms.set_calculator(calc)
            db.write(atoms, converged=True)

            # Create a new structure with exactly the same
            # composition, but higher energy
            calc = SinglePointCalculator(
                atoms, energy=cnv_hull_enegies[n_cu] + 0.5)
            atoms.set_calculator(calc)
            db.write(atoms, converged=True)

        cnv_hull = ConvexHull(db_name, conc_ranges={"Au": (0, 1)})
        energies = []
        comp = []
        for row in db.select():
            energies.append(row.energy/row.natoms)
            count = row.count_atoms()

            for k in ["Au", "Cu"]:
                if k in count.keys():
                    count[k] /= row.natoms
                else:
                    count[k] = 0.0
            comp.append(count)
        os.remove(db_name)

        # Calculate distance to the convex hull
        for c, tot_en in zip(comp, energies):
            cnv_hull.cosine_similarity_convex_hull(c, tot_en)

        # fig = cnv_hull.plot()
        # assert len(fig.get_axes()) == 1

    def test_syst_with_one_fixed_comp(self):
        db_name = "test_fixed_comp_cnv_hull.db"
        db = connect(db_name)

        # Create energies that we know are on the
        # convex hull

        for n_cu in range(6):
            atoms = bulk("Au")
            atoms = atoms*(2, 2, 2)

            atoms[0].symbol = "Zn"
            atoms[1].symbol = "Zn"

            for i in range(n_cu):
                atoms[i+2].symbol = "Cu"

            calc = SinglePointCalculator(atoms, energy=np.random.rand())
            atoms.set_calculator(calc)
            db.write(atoms, converged=True)

        cnv_hull = ConvexHull(db_name, conc_ranges={"Au": (0, 1)})

        os.remove(db_name)
        self.assertTrue(cnv_hull._unique_elem == ["Au", "Cu", "Zn"])
        # fig = cnv_hull.plot()
        # assert len(fig.get_axes()) == 1


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from clease.calculator import attach_calculator
from clease.montecarlo import SGCMonteCarlo
from clease import Concentration, CEBulk, CorrFunction
from clease.montecarlo.constraints import ConstrainElementInserts


class TestSGCMonteCarlo(unittest.TestCase):
    def test_run(self):
        db_name = 'sgc_test_aucu.db'
        conc = Concentration(basis_elements=[['Au', 'Cu']])
        setting = CEBulk(db_name=db_name, concentration=conc,
                         crystalstructure='fcc', a=4.0,
                         max_cluster_size=3, max_cluster_dia=[5.0, 4.1],
                         size=[2, 2, 2])

        atoms = setting.atoms.copy()*(3, 3, 3)
        cf = CorrFunction(setting)
        cf_scratch = cf.get_cf(setting.atoms)
        eci = {k: 0.0 for k, v in cf_scratch.items()}

        eci['c0'] = -1.0
        eci['c2_d0000_0_00'] = -0.2
        atoms = attach_calculator(setting, atoms=atoms, eci=eci)

        E = []
        for T in [5000, 2000, 1000, 500, 100]:
            mc = SGCMonteCarlo(atoms, T, symbols=['Au', 'Cu'])
            mc.run(steps=10000, chem_pot={'c1_0': -0.02})
            E.append(mc.get_thermodynamic_quantities()['energy'])

        cf_calc = atoms.get_calculator().get_cf()
        cf_scratch = cf.get_cf(atoms)

        os.remove(db_name)
        for k, v in cf_calc.items():
            self.assertAlmostEqual(v, cf_calc[k])

        # Make sure that the energies are decreasing
        for i in range(1, len(E)):
            self.assertGreaterEqual(E[i-1], E[i])

    def test_constrain_inserts(self):
        db_name = 'test_constrain_inserts.db'
        conc = Concentration(basis_elements=[['Si', 'X'], ['O', 'C']])
        setting = CEBulk(db_name=db_name, concentration=conc,
                         crystalstructure='rocksalt', a=4.0,
                         max_cluster_size=3, max_cluster_dia=[2.51, 3.0],
                         size=[2, 2, 2])
        atoms = setting.atoms.copy()*(3, 3, 3)
        cf = CorrFunction(setting)
        cf_scratch = cf.get_cf(setting.atoms)
        eci = {k: 0.0 for k, v in cf_scratch.items()}

        eci['c0'] = -1.0
        eci['c2_d0000_0_00'] = -0.2
        atoms = attach_calculator(setting, atoms=atoms, eci=eci)

        mc = SGCMonteCarlo(atoms, 100000, symbols=['Si', 'X', 'O', 'C'])

        elem_basis = [['Si', 'X'], ['O', 'C']]
        cnst = ConstrainElementInserts(atoms, setting.index_by_basis,
                                       elem_basis)
        chem_pot = {'c1_0': 0.0, 'c1_1': -0.1, 'c1_2': 0.1}
        mc.add_constraint(cnst)
        orig_symbols = [a.symbol for a in atoms]
        mc.run(steps=1000, chem_pot=chem_pot)

        new_symb = [a.symbol for a in atoms]
        os.remove(db_name)
        # Confirm that swaps have taken place
        self.assertFalse(all(map(lambda x: x[0] == x[1],
                                 zip(orig_symbols, new_symb))))

        # Check that we only have correct entries
        for allowed, basis in zip(elem_basis, setting.index_by_basis):
            for index in basis:
                self.assertTrue(atoms[index].symbol in allowed)


if __name__ == '__main__':
    unittest.main()



"""Unit tests for the corr function class."""
import os
from clease import CEBulk, CorrFunction, Concentration
import unittest


basis_elements = [["Au", "Cu", "Si"]]
concentration = Concentration(basis_elements=basis_elements)


def get_bc_setting(db_name):
    return CEBulk(crystalstructure="fcc", a=4.05, size=[4, 4, 4],
                  concentration=concentration, db_name=db_name,
                  max_cluster_size=3, max_cluster_dia=[5.73, 5.73])


def get_mic_dists(atoms, cluster):
    """Get the MIC dist."""
    dists = []
    for indx in cluster:
        dist = atoms.get_distances(indx, cluster, mic=True)
        dists.append(dist)
    return dists


class TestCorrFunc(unittest.TestCase):
    def test_trans_matrix(self):
        """Check that the MIC distance between atoms are correct."""
        db_name = "test_corrfunc_transmat.db"
        bc_setting = get_bc_setting(db_name)
        atoms = bc_setting.atoms
        tm = bc_setting.trans_matrix
        ref_dist = atoms.get_distance(0, 1, mic=True)
        for indx in range(len(atoms)):
            dist = atoms.get_distance(indx, tm[indx][1], mic=True)
            self.assertAlmostEqual(dist, ref_dist)
        os.remove(db_name)

    def test_supercell_consistency(self):
        from clease.tools import wrap_and_sort_by_position
        basis_elements = [['Li', 'X'], ['O', 'X']]
        concentration = Concentration(basis_elements=basis_elements)
        db_name_sc = "rocksalt_sc.db"
        setting = CEBulk(crystalstructure='rocksalt',
                         a=4.05,
                         size=[1, 1, 1],
                         concentration=concentration,
                         db_name=db_name_sc,
                         max_cluster_size=3,
                         max_cluster_dia=[7.0, 4.0])
        atoms = setting.atoms.copy()
        cf = CorrFunction(setting)
        cf_dict = cf.get_cf(atoms)

        atoms = wrap_and_sort_by_position(atoms*(4, 3, 2))
        cf_dict_sc = cf.get_cf(atoms)
        for k in cf_dict_sc.keys():
            self.assertAlmostEqual(cf_dict[k], cf_dict_sc[k])
        os.remove(db_name_sc)

    def test_error_message_for_non_existent_cluster(self):
        from clease.corrFunc import ClusterNotTrackedError
        basis_elements = [['Li', 'X'], ['O', 'X']]
        concentration = Concentration(basis_elements=basis_elements)
        db_name_sc = "rocksalt_sc.db"
        setting = CEBulk(crystalstructure='rocksalt',
                         a=4.05,
                         size=[1, 1, 1],
                         concentration=concentration,
                         db_name=db_name_sc,
                         max_cluster_size=3,
                         max_cluster_dia=[7.0, 4.0])
        corr = CorrFunction(setting)
        atoms = setting.atoms
        # No error should occure
        corr.get_cf_by_names(atoms, ['c3_d0000_0_000'])

        # Try a quadruplet: Have to raise error
        with self.assertRaises(ClusterNotTrackedError):
            corr.get_cf_by_names(atoms, ['c4_d0001_0_0000'])

    def tearDown(self):
        try:
            os.remove(db_name)
        except Exception:
            pass


if __name__ == '__main__':
    unittest.main()
